# event/attr.py
# Copyright (C) 2005-2024 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php

"""Attribute implementation for _Dispatch classes.

The various listener targets for a particular event class are represented
as attributes, which refer to collections of listeners to be fired off.
These collections can exist at the class level as well as at the instance
level.  An event is fired off using code like this::

    some_object.dispatch.first_connect(arg1, arg2)

Above, ``some_object.dispatch`` would be an instance of ``_Dispatch`` and
``first_connect`` is typically an instance of ``_ListenerCollection``
if event listeners are present, or ``_EmptyListener`` if none are present.

The attribute mechanics here spend effort trying to ensure listener functions
are available with a minimum of function call overhead, that unnecessary
objects aren't created (i.e. many empty per-instance listener collections),
as well as that everything is garbage collectable when owning references are
lost.  Other features such as "propagation" of listener functions across
many ``_Dispatch`` instances, "joining" of multiple ``_Dispatch`` instances,
as well as support for subclass propagation (e.g. events assigned to
``Pool`` vs. ``QueuePool``) are all implemented here.

"""
from __future__ import annotations

import collections
from itertools import chain
import threading
from types import TracebackType
import typing
from typing import Any
from typing import cast
from typing import Collection
from typing import Deque
from typing import FrozenSet
from typing import Generic
from typing import Iterator
from typing import MutableMapping
from typing import MutableSequence
from typing import NoReturn
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union
import weakref

from . import legacy
from . import registry
from .registry import _ET
from .registry import _EventKey
from .registry import _ListenerFnType
from .. import exc
from .. import util
from ..util.concurrency import AsyncAdaptedLock
from ..util.typing import Protocol

_T = TypeVar("_T", bound=Any)

if typing.TYPE_CHECKING:
    from .base import _Dispatch
    from .base import _DispatchCommon
    from .base import _HasEventsDispatch


class RefCollection(util.MemoizedSlots, Generic[_ET]):
    __slots__ = ("ref",)

    ref: weakref.ref[RefCollection[_ET]]

    def _memoized_attr_ref(self) -> weakref.ref[RefCollection[_ET]]:
        return weakref.ref(self, registry._collection_gced)


class _empty_collection(Collection[_T]):
    def append(self, element: _T) -> None:
        pass

    def appendleft(self, element: _T) -> None:
        pass

    def extend(self, other: Sequence[_T]) -> None:
        pass

    def remove(self, element: _T) -> None:
        pass

    def __contains__(self, element: Any) -> bool:
        return False

    def __iter__(self) -> Iterator[_T]:
        return iter([])

    def clear(self) -> None:
        pass

    def __len__(self) -> int:
        return 0


_ListenerFnSequenceType = Union[Deque[_T], _empty_collection[_T]]


class _ClsLevelDispatch(RefCollection[_ET]):
    """Class-level events on :class:`._Dispatch` classes."""

    __slots__ = (
        "clsname",
        "name",
        "arg_names",
        "has_kw",
        "legacy_signatures",
        "_clslevel",
        "__weakref__",
    )

    clsname: str
    name: str
    arg_names: Sequence[str]
    has_kw: bool
    legacy_signatures: MutableSequence[legacy._LegacySignatureType]
    _clslevel: MutableMapping[
        Type[_ET], _ListenerFnSequenceType[_ListenerFnType]
    ]

    def __init__(
        self,
        parent_dispatch_cls: Type[_HasEventsDispatch[_ET]],
        fn: _ListenerFnType,
    ):
        self.name = fn.__name__
        self.clsname = parent_dispatch_cls.__name__
        argspec = util.inspect_getfullargspec(fn)
        self.arg_names = argspec.args[1:]
        self.has_kw = bool(argspec.varkw)
        self.legacy_signatures = list(
            reversed(
                sorted(
                    getattr(fn, "_legacy_signatures", []), key=lambda s: s[0]
                )
            )
        )
        fn.__doc__ = legacy._augment_fn_docs(self, parent_dispatch_cls, fn)

        self._clslevel = weakref.WeakKeyDictionary()

    def _adjust_fn_spec(
        self, fn: _ListenerFnType, named: bool
    ) -> _ListenerFnType:
        if named:
            fn = self._wrap_fn_for_kw(fn)
        if self.legacy_signatures:
            try:
                argspec = util.get_callable_argspec(fn, no_self=True)
            except TypeError:
                pass
            else:
                fn = legacy._wrap_fn_for_legacy(self, fn, argspec)
        return fn

    def _wrap_fn_for_kw(self, fn: _ListenerFnType) -> _ListenerFnType:
        def wrap_kw(*args: Any, **kw: Any) -> Any:
            argdict = dict(zip(self.arg_names, args))
            argdict.update(kw)
            return fn(**argdict)

        return wrap_kw

    def _do_insert_or_append(
        self, event_key: _EventKey[_ET], is_append: bool
    ) -> None:
        target = event_key.dispatch_target
        assert isinstance(
            target, type
        ), "Class-level Event targets must be classes."
        if not getattr(target, "_sa_propagate_class_events", True):
            raise exc.InvalidRequestError(
                f"Can't assign an event directly to the {target} class"
            )

        cls: Type[_ET]

        for cls in util.walk_subclasses(target):
            if cls is not target and cls not in self._clslevel:
                self.update_subclass(cls)
            else:
                if cls not in self._clslevel:
                    self.update_subclass(cls)
                if is_append:
                    self._clslevel[cls].append(event_key._listen_fn)
                else:
                    self._clslevel[cls].appendleft(event_key._listen_fn)
        registry._stored_in_collection(event_key, self)

    def insert(self, event_key: _EventKey[_ET], propagate: bool) -> None:
        self._do_insert_or_append(event_key, is_append=False)

    def append(self, event_key: _EventKey[_ET], propagate: bool) -> None:
        self._do_insert_or_append(event_key, is_append=True)

    def update_subclass(self, target: Type[_ET]) -> None:
        if target not in self._clslevel:
            if getattr(target, "_sa_propagate_class_events", True):
                self._clslevel[target] = collections.deque()
            else:
                self._clslevel[target] = _empty_collection()

        clslevel = self._clslevel[target]
        cls: Type[_ET]
        for cls in target.__mro__[1:]:
            if cls in self._clslevel:
                clslevel.extend(
                    [fn for fn in self._clslevel[cls] if fn not in clslevel]
                )

    def remove(self, event_key: _EventKey[_ET]) -> None:
        target = event_key.dispatch_target
        cls: Type[_ET]
        for cls in util.walk_subclasses(target):
            if cls in self._clslevel:
                self._clslevel[cls].remove(event_key._listen_fn)
        registry._removed_from_collection(event_key, self)

    def clear(self) -> None:
        """Clear all class level listeners"""

        to_clear: Set[_ListenerFnType] = set()
        for dispatcher in self._clslevel.values():
            to_clear.update(dispatcher)
            dispatcher.clear()
        registry._clear(self, to_clear)

    def for_modify(self, obj: _Dispatch[_ET]) -> _ClsLevelDispatch[_ET]:
        """Return an event collection which can be modified.

        For _ClsLevelDispatch at the class level of
        a dispatcher, this returns self.

        """
        return self


class _InstanceLevelDispatch(RefCollection[_ET], Collection[_ListenerFnType]):
    __slots__ = ()

    parent: _ClsLevelDispatch[_ET]

    def _adjust_fn_spec(
        self, fn: _ListenerFnType, named: bool
    ) -> _ListenerFnType:
        return self.parent._adjust_fn_spec(fn, named)

    def __contains__(self, item: Any) -> bool:
        raise NotImplementedError()

    def __len__(self) -> int:
        raise NotImplementedError()

    def __iter__(self) -> Iterator[_ListenerFnType]:
        raise NotImplementedError()

    def __bool__(self) -> bool:
        raise NotImplementedError()

    def exec_once(self, *args: Any, **kw: Any) -> None:
        raise NotImplementedError()

    def exec_once_unless_exception(self, *args: Any, **kw: Any) -> None:
        raise NotImplementedError()

    def _exec_w_sync_on_first_run(self, *args: Any, **kw: Any) -> None:
        raise NotImplementedError()

    def __call__(self, *args: Any, **kw: Any) -> None:
        raise NotImplementedError()

    def insert(self, event_key: _EventKey[_ET], propagate: bool) -> None:
        raise NotImplementedError()

    def append(self, event_key: _EventKey[_ET], propagate: bool) -> None:
        raise NotImplementedError()

    def remove(self, event_key: _EventKey[_ET]) -> None:
        raise NotImplementedError()

    def for_modify(
        self, obj: _DispatchCommon[_ET]
    ) -> _InstanceLevelDispatch[_ET]:
        """Return an event collection which can be modified.

        For _ClsLevelDispatch at the class level of
        a dispatcher, this returns self.

        """
        return self


class _EmptyListener(_InstanceLevelDispatch[_ET]):
    """Serves as a proxy interface to the events
    served by a _ClsLevelDispatch, when there are no
    instance-level events present.

    Is replaced by _ListenerCollection when instance-level
    events are added.

    """

    __slots__ = "parent", "parent_listeners", "name"

    propagate: FrozenSet[_ListenerFnType] = frozenset()
    listeners: Tuple[()] = ()
    parent: _ClsLevelDispatch[_ET]
    parent_listeners: _ListenerFnSequenceType[_ListenerFnType]
    name: str

    def __init__(self, parent: _ClsLevelDispatch[_ET], target_cls: Type[_ET]):
        if target_cls not in parent._clslevel:
            parent.update_subclass(target_cls)
        self.parent = parent
        self.parent_listeners = parent._clslevel[target_cls]
        self.name = parent.name

    def for_modify(
        self, obj: _DispatchCommon[_ET]
    ) -> _ListenerCollection[_ET]:
        """Return an event collection which can be modified.

        For _EmptyListener at the instance level of
        a dispatcher, this generates a new
        _ListenerCollection, applies it to the instance,
        and returns it.

        """
        obj = cast("_Dispatch[_ET]", obj)

        assert obj._instance_cls is not None
        result = _ListenerCollection(self.parent, obj._instance_cls)
        if getattr(obj, self.name) is self:
            setattr(obj, self.name, result)
        else:
            assert isinstance(getattr(obj, self.name), _JoinedListener)
        return result

    def _needs_modify(self, *args: Any, **kw: Any) -> NoReturn:
        raise NotImplementedError("need to call for_modify()")

    def exec_once(self, *args: Any, **kw: Any) -> NoReturn:
        self._needs_modify(*args, **kw)

    def exec_once_unless_exception(self, *args: Any, **kw: Any) -> NoReturn:
        self._needs_modify(*args, **kw)

    def insert(self, *args: Any, **kw: Any) -> NoReturn:
        self._needs_modify(*args, **kw)

    def append(self, *args: Any, **kw: Any) -> NoReturn:
        self._needs_modify(*args, **kw)

    def remove(self, *args: Any, **kw: Any) -> NoReturn:
        self._needs_modify(*args, **kw)

    def clear(self, *args: Any, **kw: Any) -> NoReturn:
        self._needs_modify(*args, **kw)

    def __call__(self, *args: Any, **kw: Any) -> None:
        """Execute this event."""

        for fn in self.parent_listeners:
            fn(*args, **kw)

    def __contains__(self, item: Any) -> bool:
        return item in self.parent_listeners

    def __len__(self) -> int:
        return len(self.parent_listeners)

    def __iter__(self) -> Iterator[_ListenerFnType]:
        return iter(self.parent_listeners)

    def __bool__(self) -> bool:
        return bool(self.parent_listeners)


class _MutexProtocol(Protocol):
    def __enter__(self) -> bool: ...

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]: ...


class _CompoundListener(_InstanceLevelDispatch[_ET]):
    __slots__ = (
        "_exec_once_mutex",
        "_exec_once",
        "_exec_w_sync_once",
        "_is_asyncio",
    )

    _exec_once_mutex: _MutexProtocol
    parent_listeners: Collection[_ListenerFnType]
    listeners: Collection[_ListenerFnType]
    _exec_once: bool
    _exec_w_sync_once: bool

    def __init__(self, *arg: Any, **kw: Any):
        super().__init__(*arg, **kw)
        self._is_asyncio = False

    def _set_asyncio(self) -> None:
        self._is_asyncio = True

    def _memoized_attr__exec_once_mutex(self) -> _MutexProtocol:
        if self._is_asyncio:
            return AsyncAdaptedLock()
        else:
            return threading.Lock()

    def _exec_once_impl(
        self, retry_on_exception: bool, *args: Any, **kw: Any
    ) -> None:
        with self._exec_once_mutex:
            if not self._exec_once:
                try:
                    self(*args, **kw)
                    exception = False
                except:
                    exception = True
                    raise
                finally:
                    if not exception or not retry_on_exception:
                        self._exec_once = True

    def exec_once(self, *args: Any, **kw: Any) -> None:
        """Execute this event, but only if it has not been
        executed already for this collection."""

        if not self._exec_once:
            self._exec_once_impl(False, *args, **kw)

    def exec_once_unless_exception(self, *args: Any, **kw: Any) -> None:
        """Execute this event, but only if it has not been
        executed already for this collection, or was called
        by a previous exec_once_unless_exception call and
        raised an exception.

        If exec_once was already called, then this method will never run
        the callable regardless of whether it raised or not.

        .. versionadded:: 1.3.8

        """
        if not self._exec_once:
            self._exec_once_impl(True, *args, **kw)

    def _exec_w_sync_on_first_run(self, *args: Any, **kw: Any) -> None:
        """Execute this event, and use a mutex if it has not been
        executed already for this collection, or was called
        by a previous _exec_w_sync_on_first_run call and
        raised an exception.

        If _exec_w_sync_on_first_run was already called and didn't raise an
        exception, then a mutex is not used.

        .. versionadded:: 1.4.11

        """
        if not self._exec_w_sync_once:
            with self._exec_once_mutex:
                try:
                    self(*args, **kw)
                except:
                    raise
                else:
                    self._exec_w_sync_once = True
        else:
            self(*args, **kw)

    def __call__(self, *args: Any, **kw: Any) -> None:
        """Execute this event."""

        for fn in self.parent_listeners:
            fn(*args, **kw)
        for fn in self.listeners:
            fn(*args, **kw)

    def __contains__(self, item: Any) -> bool:
        return item in self.parent_listeners or item in self.listeners

    def __len__(self) -> int:
        return len(self.parent_listeners) + len(self.listeners)

    def __iter__(self) -> Iterator[_ListenerFnType]:
        return chain(self.parent_listeners, self.listeners)

    def __bool__(self) -> bool:
        return bool(self.listeners or self.parent_listeners)


class _ListenerCollection(_CompoundListener[_ET]):
    """Instance-level attributes on instances of :class:`._Dispatch`.

    Represents a collection of listeners.

    As of 0.7.9, _ListenerCollection is only first
    created via the _EmptyListener.for_modify() method.

    """

    __slots__ = (
        "parent_listeners",
        "parent",
        "name",
        "listeners",
        "propagate",
        "__weakref__",
    )

    parent_listeners: Collection[_ListenerFnType]
    parent: _ClsLevelDispatch[_ET]
    name: str
    listeners: Deque[_ListenerFnType]
    propagate: Set[_ListenerFnType]

    def __init__(self, parent: _ClsLevelDispatch[_ET], target_cls: Type[_ET]):
        super().__init__()
        if target_cls not in parent._clslevel:
            parent.update_subclass(target_cls)
        self._exec_once = False
        self._exec_w_sync_once = False
        self.parent_listeners = parent._clslevel[target_cls]
        self.parent = parent
        self.name = parent.name
        self.listeners = collections.deque()
        self.propagate = set()

    def for_modify(
        self, obj: _DispatchCommon[_ET]
    ) -> _ListenerCollection[_ET]:
        """Return an event collection which can be modified.

        For _ListenerCollection at the instance level of
        a dispatcher, this returns self.

        """
        return self

    def _update(
        self, other: _ListenerCollection[_ET], only_propagate: bool = True
    ) -> None:
        """Populate from the listeners in another :class:`_Dispatch`
        object."""
        existing_listeners = self.listeners
        existing_listener_set = set(existing_listeners)
        self.propagate.update(other.propagate)
        other_listeners = [
            l
            for l in other.listeners
            if l not in existing_listener_set
            and not only_propagate
            or l in self.propagate
        ]

        existing_listeners.extend(other_listeners)

        if other._is_asyncio:
            self._set_asyncio()

        to_associate = other.propagate.union(other_listeners)
        registry._stored_in_collection_multi(self, other, to_associate)

    def insert(self, event_key: _EventKey[_ET], propagate: bool) -> None:
        if event_key.prepend_to_list(self, self.listeners):
            if propagate:
                self.propagate.add(event_key._listen_fn)

    def append(self, event_key: _EventKey[_ET], propagate: bool) -> None:
        if event_key.append_to_list(self, self.listeners):
            if propagate:
                self.propagate.add(event_key._listen_fn)

    def remove(self, event_key: _EventKey[_ET]) -> None:
        self.listeners.remove(event_key._listen_fn)
        self.propagate.discard(event_key._listen_fn)
        registry._removed_from_collection(event_key, self)

    def clear(self) -> None:
        registry._clear(self, self.listeners)
        self.propagate.clear()
        self.listeners.clear()


class _JoinedListener(_CompoundListener[_ET]):
    __slots__ = "parent_dispatch", "name", "local", "parent_listeners"

    parent_dispatch: _DispatchCommon[_ET]
    name: str
    local: _InstanceLevelDispatch[_ET]
    parent_listeners: Collection[_ListenerFnType]

    def __init__(
        self,
        parent_dispatch: _DispatchCommon[_ET],
        name: str,
        local: _EmptyListener[_ET],
    ):
        self._exec_once = False
        self.parent_dispatch = parent_dispatch
        self.name = name
        self.local = local
        self.parent_listeners = self.local

    if not typing.TYPE_CHECKING:
        # first error, I don't really understand:
        # Signature of "listeners" incompatible with
        # supertype "_CompoundListener"  [override]
        # the name / return type are exactly the same
        # second error is getattr_isn't typed, the cast() here
        # adds too much method overhead
        @property
        def listeners(self) -> Collection[_ListenerFnType]:
            return getattr(self.parent_dispatch, self.name)

    def _adjust_fn_spec(
        self, fn: _ListenerFnType, named: bool
    ) -> _ListenerFnType:
        return self.local._adjust_fn_spec(fn, named)

    def for_modify(self, obj: _DispatchCommon[_ET]) -> _JoinedListener[_ET]:
        self.local = self.parent_listeners = self.local.for_modify(obj)
        return self

    def insert(self, event_key: _EventKey[_ET], propagate: bool) -> None:
        self.local.insert(event_key, propagate)

    def append(self, event_key: _EventKey[_ET], propagate: bool) -> None:
        self.local.append(event_key, propagate)

    def remove(self, event_key: _EventKey[_ET]) -> None:
        self.local.remove(event_key)

    def clear(self) -> None:
        raise NotImplementedError()



import json
import os
import shutil
import time
from datetime import datetime
from threading import Lock
from typing import (
    Optional,
    Dict,
    List,
)
import xml.etree.ElementTree as eTree

from application.model import (
    UpdateType,
    FiscalInfo,
)
from application.model.configuration import Configurations
from application.model.customexception import (
    DisabledFunctionality,
    ErrorCreatingBackup,
    ErrorDeletingContent,
    ErrorRecoveringFolder,
    ErrorExtractingLoaders,
    ErrorDownloadingLoaders,
    RollbackError,
    NotInsideUpdateWindow,
    ErrorValidatingOpenedUsers,
    ErrorUpdatingUserRoles,
    CannotDeletePdvWithOrders,
    ErrorDayIsOpen,
    InvalidBOHConfigurations,
)
from application.repository.api import LoaderUpdaterAPIRepository
from application.repository.pos import (
    LoaderUpdaterPOSRepository,
    VersionsPOSRepository,
    FiscalPOSRepository,
)
from application.service import (
    CommonService,
    UpdaterService,
    StaticConfigurationService,
)
from application.util import (
    LoggerUtil,
    ValidationUtil,
    settings,
)
from systools import sys_parsefilepath
from messagebus import MessageBus
from order_api import OrderService

from update_api.dto import PosUpdaterDto

GENESIS_PATH = "../../genesis"

logger = LoggerUtil.get_logger_name(__name__)


class LoaderUpdaterService(object):
    def __init__(
        self,
        configs,              # type: Configurations
        working_on_update,    # type: Lock
        api_repository,       # type: LoaderUpdaterAPIRepository
        pos_repository,       # type: LoaderUpdaterPOSRepository
        version_repository,   # type: VersionsPOSRepository,
        updater_service,      # type: UpdaterService
        message_bus,          # type: MessageBus
        fiscal_repository,    # type: FiscalPOSRepository
    ):
        # type: (...) -> None # noqa

        self.update_type = UpdateType.loader
        self.update_name = self.update_type.name
        self.enabled = configs.updaters[self.update_name].enabled
        self.local_configs = configs.updaters[self.update_name]
        self.store_id = configs.store_id

        self.working_on_update = working_on_update
        self.api_repository = api_repository
        self.pos_repository = pos_repository
        self.version_repository = version_repository
        self.updater_service = updater_service
        self.message_bus = message_bus
        self.fiscal_repository = fiscal_repository

        data_path = sys_parsefilepath("{HVDATADIR}")
        work_dir = sys_parsefilepath(configs.backup_directory)
        self.data_relative_path = "data"
        self.server_relative_path = "data/server"
        self.databases_relative_path = "data/server/databases"
        self.data_path = os.path.join(data_path, GENESIS_PATH, self.data_relative_path)
        self.server_path = os.path.join(self.data_path, "server")
        self.databases_path = os.path.join(data_path, GENESIS_PATH, self.databases_relative_path)
        default_work_path = os.path.join(work_dir, self.update_type.name)
        self.downloads_path = os.path.join(default_work_path, "downloads")
        self.backups_path = os.path.join(default_work_path, "backups")
        self.work_path = os.path.join(default_work_path, "work_path")
        self.payments_json = os.path.join(self.server_path, "bundles/payments.json")
        self.genesis_operator_levels_json = os.path.join(self.server_path, "bundles/operatorLevels.json")
        self.htdocs_path = os.path.join(data_path, 'htdocs')
        self.htdocs_kui_path = os.path.join(self.htdocs_path, 'kui')
        self.sui_static_path = os.path.join(data_path, 'htdocs/sui/static')
        self.loaders_zip = "loaders.zip"

        if not os.path.exists(self.backups_path):
            os.makedirs(self.backups_path)
        if not os.path.exists(self.downloads_path):
            os.makedirs(self.downloads_path)

    def update_loaders(
        self,                       # type: LoaderUpdaterService
        restart_hv=False,           # type: bool
        retry_on_error=False,       # type: bool
        data=None,                  # type: Optional[PosUpdaterDto]
    ):
        # type: (...) -> None

        settings.set_update_control(self.update_name, None)
        if not self.enabled:
            settings.set_update_control(self.update_name, "$UPDATE_DISABLED")
            raise DisabledFunctionality()

        rollback_necessary = False
        backup_package_zip = ''

        try:
            logger.info("Waiting Updating Loaders")
            with self.working_on_update:
                logger.info("Updating loaders")
                settings.set_update_control(self.update_name, "$UPDATING_LOADERS")
                settings.set_update_control("updating", True)

                external_update_id = None
                last_loader_update_id = 0
                obtained_date = None

                last_loader_update = self.pos_repository.get_last_loader_update()
                if last_loader_update is not None:
                    external_update_id = last_loader_update.server_update_identifier
                    last_loader_update_id = int(last_loader_update.update_id)
                    obtained_date = last_loader_update.obtained_date

                self._delete_directories()
                changed_pos_ctrl_info = self.download_and_validate_loaders()

                settings.set_update_control(self.update_name, "$LOADERS_BACKUP")
                backup_package_zip = self._create_backup()
                rollback_necessary = True

                settings.set_update_control(self.update_name, "$APPLYING_NEW_LOADERS")
                current_update_id = str(last_loader_update_id + 1)
                self.pos_repository.insert_new_loader_update(
                    update_id=current_update_id,
                    update_name="Loaders",
                    external_update_id=external_update_id,
                    obtained_date=obtained_date,
                )

                self.pos_repository.delete_update_line(
                    external_update_id=external_update_id,
                    update_id=last_loader_update_id,
                )

                self._extract_new_loaders(self.data_path)
                self._recover_folders()

                settings.set_update_control(self.update_name, "$UPDATING_USER_FUNCTIONS")
                self.update_operator_levels()

                self.updater_service.update_databases_with_json_config(
                    update_name=self.update_name,
                    settings=settings,
                )

                self.pos_repository.mark_loader_update_as_applied()

                StaticConfigurationService.update_static_loaders()
                StaticConfigurationService.run_databases_script()
                settings.set_update_control(self.update_name, "$SUCCESS_UPDATE")

                if changed_pos_ctrl_info is not None:
                    self._update_pos_ctrl_info(changed_pos_ctrl_info)

                self._update_ui_customizations()
                self._update_fiscal_id()

                rollback_necessary = False
                self.send_result_message(
                    current_update_id=current_update_id,
                )
                if restart_hv:
                    self._delete_directories()
                    logger.info("Restarting Hypervisor to apply new Loaders")
                    self.pos_repository.common_pos_repository.restart_hv(
                        pos_id=(data and data.pos_id) or "",
                        user_id=(data and data.user_id) or "",
                        event_type="POS_UPDATER_LOADER",
                    )
                    time.sleep(5)

                logger.info("Loaders update fully concluded with success!")
        except NotInsideUpdateWindow:
            settings.set_update_control(self.update_name, "$NOT_INSIDE_UPDATE_WINDOW")
            logger.info("Sleeping until enter the update window")
            raise
        except ErrorValidatingOpenedUsers:
            settings.set_update_control(self.update_name, "$UPDATE_LOADERS_ERRORVALIDATINGOPENEDUSERS")
            logger.info("Error validating loaders. Necessary to close all the operators and the day")
            raise
        except (Exception,) as ex:
            if rollback_necessary:
                with self.working_on_update:
                    self._rollback(backup_package_zip)

            if isinstance(ex, ErrorDownloadingLoaders):
                settings.set_update_control(self.update_name, ex.message)
            else:
                settings.set_update_control(self.update_name, "$UNKNOWN_ERROR")

            if not retry_on_error:
                self.pos_repository.mark_loader_update_as_error()

            logger.exception("Not expected exception on loaders update")
            raise
        finally:
            self._delete_directories()
            settings.set_update_control("updating", False)

    def download_and_validate_loaders(self):
        settings.set_update_control(self.update_name, "$GETTING_NEW_LOADERS")
        last_versions, _ = self.version_repository.get_last_versions()
        if not last_versions:
            return None

        self._download_new_loaders(last_versions=last_versions)

        self.updater_service.create_dir(self.work_path)
        self._extract_new_loaders(os.path.join(self.work_path, self.data_relative_path))

        settings.set_update_control(self.update_name, "$VALIDATING_NEW_LOADERS")
        current_pos_ctrl_dir = os.path.join(self.work_path, self.server_relative_path, 'bundles', 'posctrl')
        changed_pos_ctrl_info = None
        if os.path.exists(current_pos_ctrl_dir):
            changed_pos_ctrl_info = self._could_apply_new_loaders()

        return changed_pos_ctrl_info

    def _update_pos_ctrl_info(self, pos_ctrl_type):
        # type: (Dict) -> None

        for pos_id, value in pos_ctrl_type.items():
            user_control_type = 1 if value.get("userControlType") == "QSR" else 2
            self.pos_repository.update_pos_ctrl_info(
                pos_id=pos_id,
                user_control_type=user_control_type,
                pos_function=value.get("posFunction")
            )

    def update_operator_levels(
            self
    ):
        # type: () -> None

        try:
            if os.path.exists(self.genesis_operator_levels_json) and os.path.exists(self.sui_static_path):
                current_operator_levels = os.path.join(self.sui_static_path, 'operatorLevels.json')

                with open(self.genesis_operator_levels_json, 'rb') as f:
                    new_operator_level = json.load(f)

                has_some_configuration = sum([len(x) for x in new_operator_level["levels"].values()]) > 0
                if has_some_configuration:
                    self._create_support_menu_buttons(
                        operator_level=new_operator_level
                    )
                    self.updater_service.save_json(
                        data=new_operator_level,
                        destination_path=current_operator_levels
                    )
                else:
                    self.updater_service.delete_file(
                        path=current_operator_levels
                    )
        except (Exception,) as e:
            logger.info("Error processing operatorLevels.json: {}".format(e))
            raise ErrorUpdatingUserRoles()

    def _update_fiscal_id(
        self,   # type: LoaderUpdaterService
    ):
        # type: (...) -> None

        logger.info("Starting update fiscal id")

        current_fiscal_wrapper_dir = os.path.join(self.work_path, self.server_relative_path, "bundles", "fiscalwrapper")
        boh_fiscal_wrapper_dir = os.path.join(self.server_path, "bundles", "fiscalwrapper")
        if not os.path.exists(boh_fiscal_wrapper_dir) or not os.path.exists(current_fiscal_wrapper_dir):
            logger.error("Fiscal wrapper loaders not found")
            return

        boh_fiscal_info = self._get_fiscal_wrapper_info(fiscal_dir=boh_fiscal_wrapper_dir)
        pos_fiscal_info = self._get_fiscal_wrapper_info(fiscal_dir=current_fiscal_wrapper_dir)
        if not boh_fiscal_info or not pos_fiscal_info:
            logger.error("Fiscal info not found. BOH: {} - POS: {}".format(boh_fiscal_info, pos_fiscal_info))
            return

        if boh_fiscal_info.nf_type == pos_fiscal_info.nf_type:
            logger.info("NF type has not changed")
            return

        fiscal_id = boh_fiscal_info.initial_order_id
        logger.info("New fiscal id: {}".format(fiscal_id))

        self.fiscal_repository.set_fiscal_id(fiscal_id=fiscal_id)


    def send_result_message(
        self,                 # type: LoaderUpdaterService
        current_update_id,    # type: str
    ):
        # type: (...) -> None

        last_versions, versions_dates = self.version_repository.get_last_versions()
        body = self._get_update_notify_body(
            last_versions=last_versions,
            versions_dates=versions_dates,
        )
        success = self.api_repository.send_result_message(
            body=body,
        )
        if success:
            self.pos_repository.update_loader_notified_date(
                update_id=current_update_id,
            )

    def _get_update_notify_body(
        self,               # type: LoaderUpdaterService
        last_versions,      # type: Dict[str, str]
        versions_dates,     # type: Dict[str, str]
    ):
        # type: (...) -> Dict[str, str]

        src_version = last_versions.get(UpdateType.binaries.value).split("|")[1].replace("SRC:", "")
        src_applied_date = versions_dates.get(UpdateType.binaries.value)
        src_applied_date = datetime.strptime(src_applied_date, '%d/%m/%Y %H:%M')
        loader_applied_date = datetime.strptime(versions_dates.get(UpdateType.loader.value), '%d/%m/%Y %H:%M')
        last_loader_update = self.pos_repository.get_last_loader_update()
        server_update_identifier = last_loader_update.server_update_identifier if last_loader_update else None

        return {
            "storeId": self.store_id,
            "srcVersion": src_version,
            "srcAppliedDate": src_applied_date.isoformat(),
            "menuVersion": last_versions.get(UpdateType.catalog.value).split("_")[0],
            "loaderAppliedDate": loader_applied_date.isoformat(),
            "externalUpdateId": server_update_identifier,
        }

    @staticmethod
    def _create_support_menu_buttons(
        operator_level,     # type: Dict
    ):
        # type: (...) -> None

        rows = int(operator_level["rows"])
        support_menu = {
            "function": "SUPPORT_MENU",
            "row": rows,
            "column": 0,
            "color": "#274E13"
        }
        update_users = {
            "function": "UPDATE_USERS",
            "row": rows,
            "column": 1,
            "color": "#274E13"
        }

        operator_level["levels"]["30"] = operator_level["levels"]["10"][:]
        operator_level["levels"]["30"].extend([support_menu, update_users])

    @staticmethod
    def _validate_users(
        model  # type: eTree.XML
    ):
        # type: (...) -> None

        operators = model.find("Operators").findall("Operator")
        for operator in operators:
            status = operator.get("state")
            if status != "LOGGEDOUT":
                logger.exception("can't update loader with operator status: {}".format(status))
                raise ErrorValidatingOpenedUsers("NEED_TO_CLOSE_OPENED_USERS")

    def _could_apply_new_loaders(self):
        # type: (...) -> Dict

        boh_pos_ctrl_dir = os.path.join(self.work_path, self.server_relative_path, 'bundles', 'posctrl')
        current_pos_ctrl_dir = os.path.join(self.server_path, 'bundles', 'posctrl')
        if not os.path.exists(current_pos_ctrl_dir) or not os.path.exists(boh_pos_ctrl_dir):
            return {}

        boh_loaders = self._get_pos_ctrl_info(boh_pos_ctrl_dir)
        pos_loaders = self._get_pos_ctrl_info(current_pos_ctrl_dir)
        table_service_has_opened_day = self._table_service_has_opened_day(pos_loaders)

        order_service = OrderService(
            pos_id=pos_loaders.keys()[0],
            message_bus=self.message_bus,
        )
        pos_instances = self._get_pos_instances(order_service=order_service)
        if len(pos_instances) > len(boh_loaders.keys()):
            difference = len(pos_instances) - len(boh_loaders.keys())
            pos_ids_to_delete = pos_instances[::-1][:difference]
            self._validate_orders_db(
                pos_ids=pos_ids_to_delete,
                order_service=order_service,
            )
            self._validate_pos_ids_to_delete(pos_ids_to_delete=pos_ids_to_delete)

        changed_pos_ctrl_info = {}
        for pos_id, boh_pos_ctrl_info in boh_loaders.items():
            if pos_id == 0:
                continue

            need_validate_pos = self._need_validate_pos(
                pos_id=pos_id,
                pos_loaders=pos_loaders,
                boh_pos_ctrl_info=boh_pos_ctrl_info,
                table_service_has_opened_day=table_service_has_opened_day,
                changed_pos_ctrl_info=changed_pos_ctrl_info
            )
            if not need_validate_pos:
                continue

            model = self.pos_repository.get_model(pos_id=pos_id)
            self._validate_users(model=model)
            if self._day_is_opened_in_pos(model=model):
                logger.exception("Trying to update loader with day not closed")
                raise ErrorDayIsOpen()

        return changed_pos_ctrl_info

    @staticmethod
    def _get_pos_instances(
        order_service,    # type: OrderService
    ):
        # type: (...) -> List[int]

        order_config = order_service.get_order_config()

        return range(order_config["instances"])

    def _validate_pos_ids_to_delete(
        self,                 # type: LoaderUpdaterService
        pos_ids_to_delete,    # type: List[int]
    ):
        # type: (...) -> None

        for pos_id in pos_ids_to_delete:
            if pos_id in pos_ids_to_delete:
                model = self.pos_repository.get_model(pos_id=pos_id)
                self._validate_users(model=model)
                if self._day_is_opened_in_pos(model=model):
                    logger.exception("Error, trying to update loader with day not closed")
                    raise ErrorDayIsOpen()

    @staticmethod
    def _validate_orders_db(
        pos_ids,          # type: List[int]
        order_service,    # type: OrderService
    ):
        # type: (...) -> None

        for pos_id in pos_ids:
            orders_by_pos_id = order_service.list_orders(
                pos_id=pos_id,
                limit=1,
            )
            if len(orders_by_pos_id) > 0:
                raise CannotDeletePdvWithOrders()

    def _need_validate_pos(
        self,
        pos_id,
        pos_loaders,
        boh_pos_ctrl_info,
        table_service_has_opened_day,
        changed_pos_ctrl_info,
    ):

        if pos_id == 0:
            return False

        is_a_new_pos = pos_id not in pos_loaders
        if is_a_new_pos:
            self._validate_new_pos(
                boh_pos_ctrl_info=boh_pos_ctrl_info,
                table_service_has_opened_day=table_service_has_opened_day,
            )
            return False

        pos_pos_ctrl_info = pos_loaders[pos_id]
        same_control_type = boh_pos_ctrl_info.user_control_type == pos_pos_ctrl_info.user_control_type
        same_pos_function = boh_pos_ctrl_info.pos_function == pos_pos_ctrl_info.pos_function
        if not same_control_type or not same_pos_function:
            changed_pos_ctrl_info[pos_id] = {
                "userControlType": boh_pos_ctrl_info.user_control_type,
                "posFunction": boh_pos_ctrl_info.pos_function,
            }
            logger.info("POS{}: has a different user_control_type {} -> {}".format(
                pos_id,
                pos_pos_ctrl_info.user_control_type,
                boh_pos_ctrl_info.user_control_type
            ))

        same_pod_type = boh_pos_ctrl_info.pod_type == pos_pos_ctrl_info.pod_type
        if not same_pod_type:
            logger.info("POS{}: has a different pod_type {} -> {}".format(
                pos_id,
                pos_pos_ctrl_info.pod_type,
                boh_pos_ctrl_info.pod_type
            ))

        return not (same_control_type and same_pod_type)

    @staticmethod
    def _validate_new_pos(
        boh_pos_ctrl_info,                  # type: PosCtrlInfo
        table_service_has_opened_day,       # type: bool
    ):
        # type: (...) -> None

        if boh_pos_ctrl_info.user_control_type == "QSR":
            return

        if not table_service_has_opened_day:
            return

        logger.exception("New TS POS being added with TS day opened")
        raise ErrorDayIsOpen()

    def _table_service_has_opened_day(
        self,               # type: LoaderUpdaterService
        pos_loaders,        # type: Dict
    ):
        # type: (...) -> bool

        for pos_id, pos_ctrl_info in pos_loaders.items():
            if pos_ctrl_info.user_control_type != "TS":
                continue

            model = self.pos_repository.get_model(pos_id=pos_id)
            return self._day_is_opened_in_pos(model=model)

        return False

    @staticmethod
    def _get_pos_ctrl_info(
        pos_ctrl_dir,       # type: str
    ):
        # type: (...) -> Dict

        loaders = {}
        loaders_paths = os.listdir(pos_ctrl_dir) or []
        loaders_paths = [x for x in loaders_paths if ".cfg" in x]
        for path in loaders_paths:
            xml_loader = eTree.parse(os.path.join(pos_ctrl_dir, path))

            user_control_type = xml_loader.find(".//*[@name='{}']/string".format("userControlType"))
            user_control_type = user_control_type.text if user_control_type is not None else None

            pod_type = xml_loader.find(".//*[@name='{}']/string".format("pod"))
            pod_type = pod_type.text if pod_type is not None else None

            pos_function = xml_loader.find(".//*[@name='{}']/string".format("posFunction"))
            pos_function = pos_function.text if pos_function is not None else None

            pos_id = int(path.split("loader_")[1].split(".")[0])
            loaders[pos_id] = PosCtrlInfo(
                pos_id=pos_id,
                user_control_type=user_control_type,
                pod_type=pod_type,
                pos_function=pos_function,
            )

        return loaders

    @staticmethod
    def _get_fiscal_wrapper_info(
        fiscal_dir,     # type: str
    ):
        # type: (...) -> Optional[FiscalInfo]

        loaders_paths = os.listdir(fiscal_dir) or []
        loaders_paths = [x for x in loaders_paths if ".cfg" in x]
        find_field_name = ".//*[@name='{}']/string"
        for path in loaders_paths:
            xml_loader = eTree.parse(os.path.join(fiscal_dir, path))
            initial_state = xml_loader.find(find_field_name.format("InitialState"))
            initial_state = initial_state.text if initial_state is not None else None
            if initial_state != "automatic":
                continue

            nf_type = xml_loader.find(find_field_name.format("NFType"))
            nf_type = nf_type.text if nf_type is not None else None

            initial_order_id = xml_loader.find(find_field_name.format("initial_order_id"))
            initial_order_id = initial_order_id.text if initial_order_id is not None else 0

            return FiscalInfo(
                nf_type=nf_type,
                initial_order_id=int(initial_order_id),
            )

        return None

    def _delete_directories(self):
        self.updater_service.delete_directory(self.work_path)

    def _check_update_window(self):
        # type: () -> None

        start_time = self.local_configs.window.start_time
        end_time = self.local_configs.window.end_time
        if not ValidationUtil().is_inside_allowed_window(start_time, end_time):
            str_start_time = start_time.strftime("%H:%M:%S")
            str_end_time = end_time.strftime("%H:%M:%S")
            message = "Not inside allowed window. Start: {} / End: {}".format(str_start_time, str_end_time)
            logger.info(message)
            raise NotInsideUpdateWindow()

    def _download_new_loaders(
        self,               # type: LoaderUpdaterService
        last_versions,      # type: Dict[int, str]
    ):
        # type: (...) -> None

        try:
            logger.info("Downloading new loaders")
            file_path = os.path.join(self.downloads_path, self.loaders_zip)
            self.updater_service.create_dir(path=self.downloads_path)
            self.api_repository.download_loaders(
                file_path=file_path,
                last_versions=last_versions
            )
            logger.info("Loaders successfully downloaded")
        except InvalidBOHConfigurations:
            logger.exception("Invalid BOH configuration")
            raise
        except (Exception,) as ex:
            logger.exception("Error downloading loaders")
            raise ErrorDownloadingLoaders(ex.message)

    def _create_backup(self):
        # type: () -> str

        try:
            logger.info("Creating loaders backup")

            backup_zip_name = "loaders_{}".format(datetime.now().strftime("%Y%m%d_%H%M%S"))
            backup_package = os.path.join(self.backups_path, backup_zip_name)
            backup_package_zip = os.path.join(self.backups_path, backup_zip_name) + ".zip"

            zip_path = os.path.join(self.work_path, self.data_relative_path)
            self.updater_service.copy_directory(self.data_path, zip_path)
            self.updater_service.save_directory_zip(zip_path, backup_package)
            self.updater_service.test_zip_file(backup_package_zip)
            self._clean_backups()

            logger.info("Loaders backup successfully created!")
        except (Exception,):
            logger.exception("Error creating loaders backup")
            raise ErrorCreatingBackup()

        return backup_package_zip

    def _clean_backups(
        self,   # type: LoaderUpdaterService
    ):
        # type: (...) -> None

        list_of_files = os.listdir(self.backups_path)
        if len(list_of_files) > 5:
            full_path_files = [os.path.join(self.backups_path, x) for x in list_of_files]
            full_path_files = sorted(full_path_files, key=lambda p: os.path.getmtime(p))
            files_to_delete = full_path_files[:-5]
            for backup_file in files_to_delete:
                try:
                    CommonService.delete_file_or_directory(path=backup_file)
                except (Exception,) as e:
                    logger.exception("Failed to remove backup file {}. Error {}".format(backup_file, e))

    def _delete_content(self, path):
        # type: (str) -> None

        try:
            logger.info("Deleting path {}".format(path))

            self.updater_service.delete_directory(path)

            logger.info("Path deleted")
        except (Exception,):
            logger.exception("Error deleting path")
            raise ErrorDeletingContent()

    def _extract_new_loaders(self, path):
        # type: (str) -> None

        try:
            logger.info("Extracting loaders in data path")

            self.updater_service.create_dir(path)
            zip_path = os.path.join(self.downloads_path, self.loaders_zip)
            self.updater_service.extract_zip(zip_path, path)

            logger.info("Loaders extracted in data path")
        except (Exception,):
            logger.exception("Error extracting loaders to data path")
            raise ErrorExtractingLoaders()

    def _recover_folders(self):
        # type: () -> None

        current_folder = None
        try:
            work_path_server_folder = os.path.join(self.work_path, self.server_relative_path)
            if not os.path.exists(work_path_server_folder):
                return

            for current_folder in os.listdir(work_path_server_folder):
                if current_folder == "bundles":
                    continue

                logger.info("Recovering {} folder".format(current_folder))
                folder_path = os.path.join(self.server_path, current_folder)
                work_path_path = os.path.join(work_path_server_folder, current_folder)
                if not os.path.isdir(work_path_path):
                    continue

                self.updater_service.copy_directory(work_path_path, folder_path)
                logger.info("{} folder recovered".format(current_folder))
        except (Exception,):
            logger.exception("Error recovering {}".format(current_folder))
            raise ErrorRecoveringFolder()

    def _rollback(self, backup_package_zip):
        # type: (str) -> None

        logger.info("Starting rollback")

        retries = 0
        max_retries = 2

        while True:
            try:
                self._delete_content(self.data_path)
                self.updater_service.extract_zip(backup_package_zip, self.data_path)
            except (Exception,):
                logger.exception("Error while executing rollback")
                if retries == max_retries:
                    raise RollbackError()

                retries += 1
                continue

            logger.info("Rollback finished")
            break

    def _update_ui_customizations(
        self,
    ):
        # (...) -> None

        try:
            customization_kui_dir = os.path.join(
                self.server_path,
                "bundles",
                "customization",
                "kui",
            )
            if not os.path.isdir(customization_kui_dir):
                return

            self.updater_service.create_dir(
                path=self.htdocs_kui_path,
                delete_if_exists=False,
            )

            os_walk = os.walk(customization_kui_dir)
            for step_index, (base_dir, dirs, files) in enumerate(os_walk):
                for dir_name in dirs:
                    file_path = os.path.join(self.htdocs_kui_path, dir_name)
                    self.updater_service.create_dir(
                        path=file_path,
                        delete_if_exists=False,
                    )

                last_path_component = os.path.split(base_dir)[1]
                use_last_path_component = step_index != 0
                for file_name in files:
                    src_file = os.path.join(
                        base_dir,
                        file_name,
                    )
                    if use_last_path_component:
                        dst_file = os.path.join(
                            self.htdocs_kui_path,
                            last_path_component,
                            file_name,
                        )
                    else:
                        dst_file = os.path.join(
                            self.htdocs_kui_path,
                            file_name,
                        )
                    shutil.copy2(
                        src_file,
                        dst_file,
                    )
        except (Exception,):
            logger.exception("Error updating customization files")

    @staticmethod
    def _day_is_opened_in_pos(
        model   # type: eTree.XML
    ):
        # type: (...) -> bool

        pos_state = model.find("PosState").get("state")
        return pos_state not in ("CLOSED", "UNDEFINED")


class PosCtrlInfo(object):
    def __init__(
        self,
        pos_id,
        user_control_type,
        pod_type,
        pos_function,
    ):
        # type: (...) -> None

        self.pod_type = pod_type
        self.pos_id = pos_id
        self.user_control_type = user_control_type
        self.pos_function = pos_function




from typing import List


class OrderLines(object):
    def __init__(
        self,           # type: OrderLines
        order_id,       # type: int
        line_numbers,   # type: List[int]
    ):
        # type: (...) -> None

        self.order_id = order_id
        self.line_numbers = line_numbers



# -*- coding: utf-8 -*-
"""
The root of the greenlet package.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__all__ = [
    '__version__',
    '_C_API',

    'GreenletExit',
    'error',

    'getcurrent',
    'greenlet',

    'gettrace',
    'settrace',
]

# pylint:disable=no-name-in-module

###
# Metadata
###
__version__ = '3.1.1'
from ._greenlet import _C_API # pylint:disable=no-name-in-module

###
# Exceptions
###
from ._greenlet import GreenletExit
from ._greenlet import error

###
# greenlets
###
from ._greenlet import getcurrent
from ._greenlet import greenlet

###
# tracing
###
try:
    from ._greenlet import gettrace
    from ._greenlet import settrace
except ImportError:
    # Tracing wasn't supported.
    # XXX: The option to disable it was removed in 1.0,
    # so this branch should be dead code.
    pass

###
# Constants
# These constants aren't documented and aren't recommended.
# In 1.0, USE_GC and USE_TRACING are always true, and USE_CONTEXT_VARS
# is the same as ``sys.version_info[:2] >= 3.7``
###
from ._greenlet import GREENLET_USE_CONTEXT_VARS # pylint:disable=unused-import
from ._greenlet import GREENLET_USE_GC # pylint:disable=unused-import
from ._greenlet import GREENLET_USE_TRACING # pylint:disable=unused-import

# Controlling the use of the gc module. Provisional API for this greenlet
# implementation in 2.0.
from ._greenlet import CLOCKS_PER_SEC # pylint:disable=unused-import
from ._greenlet import enable_optional_cleanup # pylint:disable=unused-import
from ._greenlet import get_clocks_used_doing_optional_cleanup # pylint:disable=unused-import

# Other APIS in the _greenlet module are for test support.



# uid.py - functions for handling Austrian VAT numbers
#
# Copyright (C) 2012, 2013 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""UID (Umsatzsteuer-Identifikationsnummer, Austrian VAT number).

The Austrian UID is a 9-digit number that starts with a U (optionally
preceded with AT). The last digit is a check digit.

>>> validate('AT U13585627')
'U13585627'
>>> calc_check_digit('U1358562')
'7'
>>> validate('U13585626')  # incorrect check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' -./').upper().strip()
    if number.startswith('AT'):
        number = number[2:]
    return number


def calc_check_digit(number):
    """Calculate the check digit. The number passed should not have the
    check digit included."""
    return str((6 - luhn.checksum(number[1:])) % 10)


def validate(number):
    """Check if the number is a valid VAT number. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if number[:1] != 'U' or not isdigits(number[1:]):
        raise InvalidFormat()
    if len(number) != 9:
        raise InvalidLength()
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid VAT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



import logging

from sysactions import action
from typing import Optional

from ..eft import release_wait_confirmation

logger = logging.getLogger("EftActions")


@action
def callback_mobile_reprint(
    pos_id,              # type: str
    data,                # type: str
    result_code=None,    # type: Optional[str]
    message=None,        # type: Optional[str]
):
    # type: (...) -> None

    try:
        if not data or result_code != "0":
            error = "Error mobile reprint: {} | {} | {}".format(result_code, message, data)
            logger.error(error)
            release_wait_confirmation(
                pos_id=pos_id,
                status="0",
            )
            return

        logger.debug("Success mobile reprint: {0}".format(data))
        release_wait_confirmation(
            pos_id=pos_id,
            status="1",
            data=data,
        )
    except (Exception,):
        logger.exception("Error cancel transactions mobile: {0}".format(data))
        release_wait_confirmation(
            pos_id=pos_id,
            status="0",
        )



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Union

from opentelemetry import trace
from opentelemetry.context import Context
from opentelemetry.trace.span import INVALID_SPAN
from opentelemetry.util.types import Attributes


class ExemplarFilter(ABC):
    """``ExemplarFilter`` determines which measurements are eligible for becoming an
    ``Exemplar``.

    Exemplar filters are used to filter measurements before attempting to store them
    in a reservoir.

    Reference:
        https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/metrics/sdk.md#exemplarfilter
    """

    @abstractmethod
    def should_sample(
        self,
        value: Union[int, float],
        time_unix_nano: int,
        attributes: Attributes,
        context: Context,
    ) -> bool:
        """Returns whether or not a reservoir should attempt to filter a measurement.

        Args:
            value: The value of the measurement
            timestamp: A timestamp that best represents when the measurement was taken
            attributes: The complete set of measurement attributes
            context: The Context of the measurement
        """
        raise NotImplementedError(
            "ExemplarFilter.should_sample is not implemented"
        )


class AlwaysOnExemplarFilter(ExemplarFilter):
    """An ExemplarFilter which makes all measurements eligible for being an Exemplar.

    Reference:
        https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/metrics/sdk.md#alwayson
    """

    def should_sample(
        self,
        value: Union[int, float],
        time_unix_nano: int,
        attributes: Attributes,
        context: Context,
    ) -> bool:
        """Returns whether or not a reservoir should attempt to filter a measurement.

        Args:
            value: The value of the measurement
            timestamp: A timestamp that best represents when the measurement was taken
            attributes: The complete set of measurement attributes
            context: The Context of the measurement
        """
        return True


class AlwaysOffExemplarFilter(ExemplarFilter):
    """An ExemplarFilter which makes no measurements eligible for being an Exemplar.

    Using this ExemplarFilter is as good as disabling Exemplar feature.

    Reference:
        https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/metrics/sdk.md#alwaysoff
    """

    def should_sample(
        self,
        value: Union[int, float],
        time_unix_nano: int,
        attributes: Attributes,
        context: Context,
    ) -> bool:
        """Returns whether or not a reservoir should attempt to filter a measurement.

        Args:
            value: The value of the measurement
            timestamp: A timestamp that best represents when the measurement was taken
            attributes: The complete set of measurement attributes
            context: The Context of the measurement
        """
        return False


class TraceBasedExemplarFilter(ExemplarFilter):
    """An ExemplarFilter which makes those measurements eligible for being an Exemplar,
    which are recorded in the context of a sampled parent span.

    Reference:
        https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/metrics/sdk.md#tracebased
    """

    def should_sample(
        self,
        value: Union[int, float],
        time_unix_nano: int,
        attributes: Attributes,
        context: Context,
    ) -> bool:
        """Returns whether or not a reservoir should attempt to filter a measurement.

        Args:
            value: The value of the measurement
            timestamp: A timestamp that best represents when the measurement was taken
            attributes: The complete set of measurement attributes
            context: The Context of the measurement
        """
        span = trace.get_current_span(context)
        if span == INVALID_SPAN:
            return False
        return span.get_span_context().trace_flags.sampled



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from logging import getLogger
from typing import Collection

from packaging.requirements import InvalidRequirement, Requirement

from opentelemetry.util._importlib_metadata import (
    Distribution,
    PackageNotFoundError,
    version,
)

logger = getLogger(__name__)


class DependencyConflict:
    required: str | None = None
    found: str | None = None

    def __init__(self, required: str | None, found: str | None = None):
        self.required = required
        self.found = found

    def __str__(self):
        return f'DependencyConflict: requested: "{self.required}" but found: "{self.found}"'


def get_dist_dependency_conflicts(
    dist: Distribution,
) -> DependencyConflict | None:
    instrumentation_deps = []
    extra = "extra"
    instruments = "instruments"
    instruments_marker = {extra: instruments}
    for dep in dist.requires:
        if extra not in dep or instruments not in dep:
            continue

        req = Requirement(dep)
        if req.marker.evaluate(instruments_marker):
            instrumentation_deps.append(req)

    return get_dependency_conflicts(instrumentation_deps)


def get_dependency_conflicts(
    deps: Collection[str | Requirement],
) -> DependencyConflict | None:
    for dep in deps:
        if isinstance(dep, Requirement):
            req = dep
        else:
            try:
                req = Requirement(dep)
            except InvalidRequirement as exc:
                logger.warning(
                    'error parsing dependency, reporting as a conflict: "%s" - %s',
                    dep,
                    exc,
                )
                return DependencyConflict(dep)

        try:
            dist_version = version(req.name)
        except PackageNotFoundError:
            return DependencyConflict(dep)

        if not req.specifier.contains(dist_version):
            return DependencyConflict(dep, f"{req.name} {dist_version}")
    return None



from xml.etree import cElementTree as eTree

from msgbus import (
    MBEasyContext,
    TK_POS_GETMODEL,
    FM_PARAM,
    TK_SYS_ACK,
)


class PosModelRetriever(object):
    def __init__(
        self,
        mb_context,  # type: MBEasyContext
    ):
        self.mb_context = mb_context

    def get_model(
        self,
        pos_id,  # type: int
    ):
        reply = self.mb_context.MB_EasySendMessage(
            dest_name="POS{}".format(pos_id),
            token=TK_POS_GETMODEL,
            format=FM_PARAM,
            data="{}".format(pos_id),
        )
        if reply.token != TK_SYS_ACK:
            raise Exception("Could not retrieve state of pos '{}'".format(pos_id))
        return eTree.XML(reply.data)




from greenlet import greenlet

from . import TestCase

class genlet(greenlet):
    parent = None
    def __init__(self, *args, **kwds):
        self.args = args
        self.kwds = kwds

    def run(self):
        fn, = self.fn
        fn(*self.args, **self.kwds)

    def __iter__(self):
        return self

    def __next__(self):
        self.parent = greenlet.getcurrent()
        result = self.switch()
        if self:
            return result

        raise StopIteration

    next = __next__


def Yield(value):
    g = greenlet.getcurrent()
    while not isinstance(g, genlet):
        if g is None:
            raise RuntimeError('yield outside a genlet')
        g = g.parent
    g.parent.switch(value)


def generator(func):
    class Generator(genlet):
        fn = (func,)
    return Generator

# ____________________________________________________________


class GeneratorTests(TestCase):
    def test_generator(self):
        seen = []

        def g(n):
            for i in range(n):
                seen.append(i)
                Yield(i)
        g = generator(g)
        for _ in range(3):
            for j in g(5):
                seen.append(j)
        self.assertEqual(seen, 3 * [0, 0, 1, 1, 2, 2, 3, 3, 4, 4])



import json


class PosUpdaterDto(object):

    def __init__(
        self,       # type: PosUpdaterDto
        pos_id,     # type: int
        user_id,    # type: str
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.user_id = user_id

    def __str__(
        self,   # type: PosUpdaterDto
    ):
        # type: (...) -> str

        return json.dumps({
            "pos_id": self.pos_id,
            "user_id": self.user_id,
        })



import logging
from os import PathLike
from typing import BinaryIO, List, Optional, Set, Union

from .cd import (
    coherence_ratio,
    encoding_languages,
    mb_encoding_languages,
    merge_coherence_ratios,
)
from .constant import IANA_SUPPORTED, TOO_BIG_SEQUENCE, TOO_SMALL_SEQUENCE, TRACE
from .md import mess_ratio
from .models import CharsetMatch, CharsetMatches
from .utils import (
    any_specified_encoding,
    cut_sequence_chunks,
    iana_name,
    identify_sig_or_bom,
    is_cp_similar,
    is_multi_byte_encoding,
    should_strip_sig_or_bom,
)

# Will most likely be controversial
# logging.addLevelName(TRACE, "TRACE")
logger = logging.getLogger("charset_normalizer")
explain_handler = logging.StreamHandler()
explain_handler.setFormatter(
    logging.Formatter("%(asctime)s | %(levelname)s | %(message)s")
)


def from_bytes(
    sequences: Union[bytes, bytearray],
    steps: int = 5,
    chunk_size: int = 512,
    threshold: float = 0.2,
    cp_isolation: Optional[List[str]] = None,
    cp_exclusion: Optional[List[str]] = None,
    preemptive_behaviour: bool = True,
    explain: bool = False,
    language_threshold: float = 0.1,
    enable_fallback: bool = True,
) -> CharsetMatches:
    """
    Given a raw bytes sequence, return the best possibles charset usable to render str objects.
    If there is no results, it is a strong indicator that the source is binary/not text.
    By default, the process will extract 5 blocks of 512o each to assess the mess and coherence of a given sequence.
    And will give up a particular code page after 20% of measured mess. Those criteria are customizable at will.

    The preemptive behavior DOES NOT replace the traditional detection workflow, it prioritize a particular code page
    but never take it for granted. Can improve the performance.

    You may want to focus your attention to some code page or/and not others, use cp_isolation and cp_exclusion for that
    purpose.

    This function will strip the SIG in the payload/sequence every time except on UTF-16, UTF-32.
    By default the library does not setup any handler other than the NullHandler, if you choose to set the 'explain'
    toggle to True it will alter the logger configuration to add a StreamHandler that is suitable for debugging.
    Custom logging format and handler can be set manually.
    """

    if not isinstance(sequences, (bytearray, bytes)):
        raise TypeError(
            "Expected object of type bytes or bytearray, got: {0}".format(
                type(sequences)
            )
        )

    if explain:
        previous_logger_level: int = logger.level
        logger.addHandler(explain_handler)
        logger.setLevel(TRACE)

    length: int = len(sequences)

    if length == 0:
        logger.debug("Encoding detection on empty bytes, assuming utf_8 intention.")
        if explain:
            logger.removeHandler(explain_handler)
            logger.setLevel(previous_logger_level or logging.WARNING)
        return CharsetMatches([CharsetMatch(sequences, "utf_8", 0.0, False, [], "")])

    if cp_isolation is not None:
        logger.log(
            TRACE,
            "cp_isolation is set. use this flag for debugging purpose. "
            "limited list of encoding allowed : %s.",
            ", ".join(cp_isolation),
        )
        cp_isolation = [iana_name(cp, False) for cp in cp_isolation]
    else:
        cp_isolation = []

    if cp_exclusion is not None:
        logger.log(
            TRACE,
            "cp_exclusion is set. use this flag for debugging purpose. "
            "limited list of encoding excluded : %s.",
            ", ".join(cp_exclusion),
        )
        cp_exclusion = [iana_name(cp, False) for cp in cp_exclusion]
    else:
        cp_exclusion = []

    if length <= (chunk_size * steps):
        logger.log(
            TRACE,
            "override steps (%i) and chunk_size (%i) as content does not fit (%i byte(s) given) parameters.",
            steps,
            chunk_size,
            length,
        )
        steps = 1
        chunk_size = length

    if steps > 1 and length / steps < chunk_size:
        chunk_size = int(length / steps)

    is_too_small_sequence: bool = len(sequences) < TOO_SMALL_SEQUENCE
    is_too_large_sequence: bool = len(sequences) >= TOO_BIG_SEQUENCE

    if is_too_small_sequence:
        logger.log(
            TRACE,
            "Trying to detect encoding from a tiny portion of ({}) byte(s).".format(
                length
            ),
        )
    elif is_too_large_sequence:
        logger.log(
            TRACE,
            "Using lazy str decoding because the payload is quite large, ({}) byte(s).".format(
                length
            ),
        )

    prioritized_encodings: List[str] = []

    specified_encoding: Optional[str] = (
        any_specified_encoding(sequences) if preemptive_behaviour else None
    )

    if specified_encoding is not None:
        prioritized_encodings.append(specified_encoding)
        logger.log(
            TRACE,
            "Detected declarative mark in sequence. Priority +1 given for %s.",
            specified_encoding,
        )

    tested: Set[str] = set()
    tested_but_hard_failure: List[str] = []
    tested_but_soft_failure: List[str] = []

    fallback_ascii: Optional[CharsetMatch] = None
    fallback_u8: Optional[CharsetMatch] = None
    fallback_specified: Optional[CharsetMatch] = None

    results: CharsetMatches = CharsetMatches()

    early_stop_results: CharsetMatches = CharsetMatches()

    sig_encoding, sig_payload = identify_sig_or_bom(sequences)

    if sig_encoding is not None:
        prioritized_encodings.append(sig_encoding)
        logger.log(
            TRACE,
            "Detected a SIG or BOM mark on first %i byte(s). Priority +1 given for %s.",
            len(sig_payload),
            sig_encoding,
        )

    prioritized_encodings.append("ascii")

    if "utf_8" not in prioritized_encodings:
        prioritized_encodings.append("utf_8")

    for encoding_iana in prioritized_encodings + IANA_SUPPORTED:
        if cp_isolation and encoding_iana not in cp_isolation:
            continue

        if cp_exclusion and encoding_iana in cp_exclusion:
            continue

        if encoding_iana in tested:
            continue

        tested.add(encoding_iana)

        decoded_payload: Optional[str] = None
        bom_or_sig_available: bool = sig_encoding == encoding_iana
        strip_sig_or_bom: bool = bom_or_sig_available and should_strip_sig_or_bom(
            encoding_iana
        )

        if encoding_iana in {"utf_16", "utf_32"} and not bom_or_sig_available:
            logger.log(
                TRACE,
                "Encoding %s won't be tested as-is because it require a BOM. Will try some sub-encoder LE/BE.",
                encoding_iana,
            )
            continue
        if encoding_iana in {"utf_7"} and not bom_or_sig_available:
            logger.log(
                TRACE,
                "Encoding %s won't be tested as-is because detection is unreliable without BOM/SIG.",
                encoding_iana,
            )
            continue

        try:
            is_multi_byte_decoder: bool = is_multi_byte_encoding(encoding_iana)
        except (ModuleNotFoundError, ImportError):
            logger.log(
                TRACE,
                "Encoding %s does not provide an IncrementalDecoder",
                encoding_iana,
            )
            continue

        try:
            if is_too_large_sequence and is_multi_byte_decoder is False:
                str(
                    (
                        sequences[: int(50e4)]
                        if strip_sig_or_bom is False
                        else sequences[len(sig_payload) : int(50e4)]
                    ),
                    encoding=encoding_iana,
                )
            else:
                decoded_payload = str(
                    (
                        sequences
                        if strip_sig_or_bom is False
                        else sequences[len(sig_payload) :]
                    ),
                    encoding=encoding_iana,
                )
        except (UnicodeDecodeError, LookupError) as e:
            if not isinstance(e, LookupError):
                logger.log(
                    TRACE,
                    "Code page %s does not fit given bytes sequence at ALL. %s",
                    encoding_iana,
                    str(e),
                )
            tested_but_hard_failure.append(encoding_iana)
            continue

        similar_soft_failure_test: bool = False

        for encoding_soft_failed in tested_but_soft_failure:
            if is_cp_similar(encoding_iana, encoding_soft_failed):
                similar_soft_failure_test = True
                break

        if similar_soft_failure_test:
            logger.log(
                TRACE,
                "%s is deemed too similar to code page %s and was consider unsuited already. Continuing!",
                encoding_iana,
                encoding_soft_failed,
            )
            continue

        r_ = range(
            0 if not bom_or_sig_available else len(sig_payload),
            length,
            int(length / steps),
        )

        multi_byte_bonus: bool = (
            is_multi_byte_decoder
            and decoded_payload is not None
            and len(decoded_payload) < length
        )

        if multi_byte_bonus:
            logger.log(
                TRACE,
                "Code page %s is a multi byte encoding table and it appear that at least one character "
                "was encoded using n-bytes.",
                encoding_iana,
            )

        max_chunk_gave_up: int = int(len(r_) / 4)

        max_chunk_gave_up = max(max_chunk_gave_up, 2)
        early_stop_count: int = 0
        lazy_str_hard_failure = False

        md_chunks: List[str] = []
        md_ratios = []

        try:
            for chunk in cut_sequence_chunks(
                sequences,
                encoding_iana,
                r_,
                chunk_size,
                bom_or_sig_available,
                strip_sig_or_bom,
                sig_payload,
                is_multi_byte_decoder,
                decoded_payload,
            ):
                md_chunks.append(chunk)

                md_ratios.append(
                    mess_ratio(
                        chunk,
                        threshold,
                        explain is True and 1 <= len(cp_isolation) <= 2,
                    )
                )

                if md_ratios[-1] >= threshold:
                    early_stop_count += 1

                if (early_stop_count >= max_chunk_gave_up) or (
                    bom_or_sig_available and strip_sig_or_bom is False
                ):
                    break
        except (
            UnicodeDecodeError
        ) as e:  # Lazy str loading may have missed something there
            logger.log(
                TRACE,
                "LazyStr Loading: After MD chunk decode, code page %s does not fit given bytes sequence at ALL. %s",
                encoding_iana,
                str(e),
            )
            early_stop_count = max_chunk_gave_up
            lazy_str_hard_failure = True

        # We might want to check the sequence again with the whole content
        # Only if initial MD tests passes
        if (
            not lazy_str_hard_failure
            and is_too_large_sequence
            and not is_multi_byte_decoder
        ):
            try:
                sequences[int(50e3) :].decode(encoding_iana, errors="strict")
            except UnicodeDecodeError as e:
                logger.log(
                    TRACE,
                    "LazyStr Loading: After final lookup, code page %s does not fit given bytes sequence at ALL. %s",
                    encoding_iana,
                    str(e),
                )
                tested_but_hard_failure.append(encoding_iana)
                continue

        mean_mess_ratio: float = sum(md_ratios) / len(md_ratios) if md_ratios else 0.0
        if mean_mess_ratio >= threshold or early_stop_count >= max_chunk_gave_up:
            tested_but_soft_failure.append(encoding_iana)
            logger.log(
                TRACE,
                "%s was excluded because of initial chaos probing. Gave up %i time(s). "
                "Computed mean chaos is %f %%.",
                encoding_iana,
                early_stop_count,
                round(mean_mess_ratio * 100, ndigits=3),
            )
            # Preparing those fallbacks in case we got nothing.
            if (
                enable_fallback
                and encoding_iana in ["ascii", "utf_8", specified_encoding]
                and not lazy_str_hard_failure
            ):
                fallback_entry = CharsetMatch(
                    sequences,
                    encoding_iana,
                    threshold,
                    False,
                    [],
                    decoded_payload,
                    preemptive_declaration=specified_encoding,
                )
                if encoding_iana == specified_encoding:
                    fallback_specified = fallback_entry
                elif encoding_iana == "ascii":
                    fallback_ascii = fallback_entry
                else:
                    fallback_u8 = fallback_entry
            continue

        logger.log(
            TRACE,
            "%s passed initial chaos probing. Mean measured chaos is %f %%",
            encoding_iana,
            round(mean_mess_ratio * 100, ndigits=3),
        )

        if not is_multi_byte_decoder:
            target_languages: List[str] = encoding_languages(encoding_iana)
        else:
            target_languages = mb_encoding_languages(encoding_iana)

        if target_languages:
            logger.log(
                TRACE,
                "{} should target any language(s) of {}".format(
                    encoding_iana, str(target_languages)
                ),
            )

        cd_ratios = []

        # We shall skip the CD when its about ASCII
        # Most of the time its not relevant to run "language-detection" on it.
        if encoding_iana != "ascii":
            for chunk in md_chunks:
                chunk_languages = coherence_ratio(
                    chunk,
                    language_threshold,
                    ",".join(target_languages) if target_languages else None,
                )

                cd_ratios.append(chunk_languages)

        cd_ratios_merged = merge_coherence_ratios(cd_ratios)

        if cd_ratios_merged:
            logger.log(
                TRACE,
                "We detected language {} using {}".format(
                    cd_ratios_merged, encoding_iana
                ),
            )

        current_match = CharsetMatch(
            sequences,
            encoding_iana,
            mean_mess_ratio,
            bom_or_sig_available,
            cd_ratios_merged,
            (
                decoded_payload
                if (
                    is_too_large_sequence is False
                    or encoding_iana in [specified_encoding, "ascii", "utf_8"]
                )
                else None
            ),
            preemptive_declaration=specified_encoding,
        )

        results.append(current_match)

        if (
            encoding_iana in [specified_encoding, "ascii", "utf_8"]
            and mean_mess_ratio < 0.1
        ):
            # If md says nothing to worry about, then... stop immediately!
            if mean_mess_ratio == 0.0:
                logger.debug(
                    "Encoding detection: %s is most likely the one.",
                    current_match.encoding,
                )
                if explain:
                    logger.removeHandler(explain_handler)
                    logger.setLevel(previous_logger_level)
                return CharsetMatches([current_match])

            early_stop_results.append(current_match)

        if (
            len(early_stop_results)
            and (specified_encoding is None or specified_encoding in tested)
            and "ascii" in tested
            and "utf_8" in tested
        ):
            probable_result: CharsetMatch = early_stop_results.best()  # type: ignore[assignment]
            logger.debug(
                "Encoding detection: %s is most likely the one.",
                probable_result.encoding,
            )
            if explain:
                logger.removeHandler(explain_handler)
                logger.setLevel(previous_logger_level)

            return CharsetMatches([probable_result])

        if encoding_iana == sig_encoding:
            logger.debug(
                "Encoding detection: %s is most likely the one as we detected a BOM or SIG within "
                "the beginning of the sequence.",
                encoding_iana,
            )
            if explain:
                logger.removeHandler(explain_handler)
                logger.setLevel(previous_logger_level)
            return CharsetMatches([results[encoding_iana]])

    if len(results) == 0:
        if fallback_u8 or fallback_ascii or fallback_specified:
            logger.log(
                TRACE,
                "Nothing got out of the detection process. Using ASCII/UTF-8/Specified fallback.",
            )

        if fallback_specified:
            logger.debug(
                "Encoding detection: %s will be used as a fallback match",
                fallback_specified.encoding,
            )
            results.append(fallback_specified)
        elif (
            (fallback_u8 and fallback_ascii is None)
            or (
                fallback_u8
                and fallback_ascii
                and fallback_u8.fingerprint != fallback_ascii.fingerprint
            )
            or (fallback_u8 is not None)
        ):
            logger.debug("Encoding detection: utf_8 will be used as a fallback match")
            results.append(fallback_u8)
        elif fallback_ascii:
            logger.debug("Encoding detection: ascii will be used as a fallback match")
            results.append(fallback_ascii)

    if results:
        logger.debug(
            "Encoding detection: Found %s as plausible (best-candidate) for content. With %i alternatives.",
            results.best().encoding,  # type: ignore
            len(results) - 1,
        )
    else:
        logger.debug("Encoding detection: Unable to determine any suitable charset.")

    if explain:
        logger.removeHandler(explain_handler)
        logger.setLevel(previous_logger_level)

    return results


def from_fp(
    fp: BinaryIO,
    steps: int = 5,
    chunk_size: int = 512,
    threshold: float = 0.20,
    cp_isolation: Optional[List[str]] = None,
    cp_exclusion: Optional[List[str]] = None,
    preemptive_behaviour: bool = True,
    explain: bool = False,
    language_threshold: float = 0.1,
    enable_fallback: bool = True,
) -> CharsetMatches:
    """
    Same thing than the function from_bytes but using a file pointer that is already ready.
    Will not close the file pointer.
    """
    return from_bytes(
        fp.read(),
        steps,
        chunk_size,
        threshold,
        cp_isolation,
        cp_exclusion,
        preemptive_behaviour,
        explain,
        language_threshold,
        enable_fallback,
    )


def from_path(
    path: Union[str, bytes, PathLike],  # type: ignore[type-arg]
    steps: int = 5,
    chunk_size: int = 512,
    threshold: float = 0.20,
    cp_isolation: Optional[List[str]] = None,
    cp_exclusion: Optional[List[str]] = None,
    preemptive_behaviour: bool = True,
    explain: bool = False,
    language_threshold: float = 0.1,
    enable_fallback: bool = True,
) -> CharsetMatches:
    """
    Same thing than the function from_bytes but with one extra step. Opening and reading given file path in binary mode.
    Can raise IOError.
    """
    with open(path, "rb") as fp:
        return from_fp(
            fp,
            steps,
            chunk_size,
            threshold,
            cp_isolation,
            cp_exclusion,
            preemptive_behaviour,
            explain,
            language_threshold,
            enable_fallback,
        )


def is_binary(
    fp_or_path_or_payload: Union[PathLike, str, BinaryIO, bytes],  # type: ignore[type-arg]
    steps: int = 5,
    chunk_size: int = 512,
    threshold: float = 0.20,
    cp_isolation: Optional[List[str]] = None,
    cp_exclusion: Optional[List[str]] = None,
    preemptive_behaviour: bool = True,
    explain: bool = False,
    language_threshold: float = 0.1,
    enable_fallback: bool = False,
) -> bool:
    """
    Detect if the given input (file, bytes, or path) points to a binary file. aka. not a string.
    Based on the same main heuristic algorithms and default kwargs at the sole exception that fallbacks match
    are disabled to be stricter around ASCII-compatible but unlikely to be a string.
    """
    if isinstance(fp_or_path_or_payload, (str, PathLike)):
        guesses = from_path(
            fp_or_path_or_payload,
            steps=steps,
            chunk_size=chunk_size,
            threshold=threshold,
            cp_isolation=cp_isolation,
            cp_exclusion=cp_exclusion,
            preemptive_behaviour=preemptive_behaviour,
            explain=explain,
            language_threshold=language_threshold,
            enable_fallback=enable_fallback,
        )
    elif isinstance(
        fp_or_path_or_payload,
        (
            bytes,
            bytearray,
        ),
    ):
        guesses = from_bytes(
            fp_or_path_or_payload,
            steps=steps,
            chunk_size=chunk_size,
            threshold=threshold,
            cp_isolation=cp_isolation,
            cp_exclusion=cp_exclusion,
            preemptive_behaviour=preemptive_behaviour,
            explain=explain,
            language_threshold=language_threshold,
            enable_fallback=enable_fallback,
        )
    else:
        guesses = from_fp(
            fp_or_path_or_payload,
            steps=steps,
            chunk_size=chunk_size,
            threshold=threshold,
            cp_isolation=cp_isolation,
            cp_exclusion=cp_exclusion,
            preemptive_behaviour=preemptive_behaviour,
            explain=explain,
            language_threshold=language_threshold,
            enable_fallback=enable_fallback,
        )

    return not guesses



import json
from typing import (
    List,
    Optional,
    Union,
)
from xml.etree import ElementTree as eTree

from boss_api.shared import token
from boss_api.shared.exception import (
    AccessDenied,
    AuthenticationFailed,
    BossException,
)
from boss_api.shared.response import UserDto
from messagebus import (
    ApplicationErrorMessage,
    BadRequestMessage,
    DataType,
    DefaultToken,
    Message,
    MessageBus,
)


class Boss(object):

    def __init__(
        self,           # type: Boss
        message_bus,    # type: MessageBus
    ):
        # type: (...) -> None

        self.message_bus = message_bus

    def authenticate(
        self,               # type: Boss
        user_id,            # type: int
        passwords,          # type: Union[str, List[str]]
        min_level=None,     # type: Optional[int]
    ):
        # type: (...) -> UserDto

        if isinstance(passwords, list):
            passwords = "|".join(passwords)

        request = "{}\0{}".format(user_id or "", passwords or "").encode("utf-8")

        reply = self.message_bus.send_message(
            component="UserControl",
            message=Message(
                token=token.TK_USER_CONTROL_AUTHENTICATE,
                data_type=DataType.param,
                data=request,
            )
        )

        if reply.token == DefaultToken.TK_SYS_ACK and reply.data is not None:
            response = reply.data.decode("utf-8")
            fields = response.split("\0")

            user = UserDto(
                user_id=user_id,
                level=fields[0],
                user_name=fields[1],
                long_name=fields[2],
                status=fields[3],
                admission_date=fields[4],
                termination_date=fields[5],
            )
            if min_level is not None and user.level < min_level:
                raise AccessDenied()

            return user

        if reply.token == DefaultToken.TK_SYS_NAK:
            raise AuthenticationFailed()

        raise self._build_exception_from_reply(reply=reply)

    def authenticate_nfc(
        self,               # type: Boss
        nfc_uuid,           # type: str
        min_level=None,     # type: Optional[int]
    ):
        # type: (...) -> UserDto

        request = str(nfc_uuid).encode("utf-8")

        reply = self.message_bus.send_message(
            component="UserControl",
            message=Message(
                token=token.TK_USER_CONTROL_AUTHENTICATE_NFC,
                data_type=DataType.param,
                data=request,
            )
        )

        if reply.token == DefaultToken.TK_SYS_ACK and reply.data is not None:
            response = reply.data.decode("utf-8")
            fields = response.split("\0")

            user = UserDto(
                user_id=fields[0],
                level=fields[1],
                user_name=fields[2],
                long_name=fields[3],
                status=fields[4],
                admission_date=fields[5],
                termination_date=fields[6],
            )
            if min_level is not None and user.level < min_level:
                raise AccessDenied()

            return user

        if reply.token == DefaultToken.TK_SYS_NAK:
            raise AuthenticationFailed()

        raise self._build_exception_from_reply(reply=reply)

    def get_info(
        self,       # type: Boss
        user_id,    # type: int
    ):
        # type: (...) -> UserDto

        request = str(user_id).encode("utf-8")

        reply = self.message_bus.send_message(
            component="UserControl",
            message=Message(
                token=token.TK_USER_CONTROL_GET_INFO,
                data_type=DataType.param,
                data=request,
            )
        )

        if reply.token == DefaultToken.TK_SYS_ACK and reply.data is not None:
            response = eTree.fromstring(reply.data.decode("utf-8"))
            user = response.find("user")
            return UserDto(
                user_id=user_id,
                level=int(user.get("Level")),
                user_name=user.get("UserName"),
                long_name=user.get("LongName"),
                status=int(user.get("Status")),
                admission_date=user.get("AdmissionDate"),
                termination_date=user.get("TerminationDate"),
            )

        if reply.token == DefaultToken.TK_SYS_NAK:
            raise AuthenticationFailed()

        raise self._build_exception_from_reply(reply=reply)

    def get_users_list(
        self,       # type: Boss
    ):
        # type: (...) -> List[UserDto]

        reply = self.message_bus.send_message(
            component="UserControl",
            message=Message(token=token.TK_USER_CONTROL_GET_USERS_LIST)
        )

        if reply.token == DefaultToken.TK_SYS_ACK and reply.data is not None:
            response = eTree.fromstring(reply.data.decode("utf-8"))
            users = response.findall("user")
            users_list = []
            for user_data in users:
                users_list.append(
                    UserDto(
                        user_id=int(user_data.get("UserId")),
                        level=int(user_data.get("Level")),
                        user_name=user_data.get("UserName"),
                        long_name=user_data.get("LongName"),
                        status=int(user_data.get("Status")),
                        admission_date=user_data.get("AdmissionDate"),
                        termination_date=user_data.get("TerminationDate"),
                    )
                )

            return users_list

        if reply.token == DefaultToken.TK_SYS_NAK:
            raise AuthenticationFailed()

        raise self._build_exception_from_reply(reply=reply)

    def insert_or_update_users(
        self,       # type: Boss
        users,      # type: List[UserDto]
    ):
        # type: (...) -> None

        users_json = []
        for user in users:
            users_json.append(
                {
                    "user_id": user.user_id,
                    "user_name": user.user_name,
                    "long_name": user.long_name,
                    "password": user.password,
                    "level": user.level,
                    "status": user.status,
                    "admission_date": user.admission_date,
                    "termination_date": user.termination_date,
                    "fingerprint_minutiae": user.fingerprint_minutiae,
                    "nfc_uuid": user.nfc_uuid,
                }
            )

        reply = self.message_bus.send_message(
            component="UserControl",
            message=Message(
                token=token.TK_USER_CONTROL_INSERT_OR_UPDATE_USERS,
                data_type=DataType.json,
                data=json.dumps(users_json).encode("utf-8"),
            ),
        )

        if reply.token == DefaultToken.TK_SYS_ACK:
            return

        raise self._build_exception_from_reply(reply=reply)

    def _build_exception_from_reply(
        self,   # type: Boss
        reply,  # type: Message
    ):
        # type: (...) -> Exception

        if reply.token == DefaultToken.TK_SYS_BAD_REQUEST:
            bad_request = BadRequestMessage.build_from_message(message=reply)
            exception = BossException.get_exception(error_code=bad_request.error_code)
            return exception

        elif reply.token == DefaultToken.TK_SYS_APPLICATION_ERROR:
            application_error = ApplicationErrorMessage.build_from_message(message=reply)
            exception = BossException.get_exception(error_code=application_error.error_code)
            exception.error_message = application_error.error_message
            return exception

        return self._build_internal_error(message=reply)

    @staticmethod
    def _build_internal_error(
        message,  # type: Message
    ):
        # type: (...) -> RuntimeError

        try:
            data_text = message.data.decode("utf-8")
        except UnicodeDecodeError:
            data_text = message.data.hex()

        return RuntimeError("Invalid response: {} - {}".format(message.token, data_text))



# util/topological.py
# Copyright (C) 2005-2025 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php

"""Topological sorting algorithms."""

from __future__ import annotations

from typing import Any
from typing import Collection
from typing import DefaultDict
from typing import Iterable
from typing import Iterator
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import TypeVar

from .. import util
from ..exc import CircularDependencyError

_T = TypeVar("_T", bound=Any)

__all__ = ["sort", "sort_as_subsets", "find_cycles"]


def sort_as_subsets(
    tuples: Collection[Tuple[_T, _T]], allitems: Collection[_T]
) -> Iterator[Sequence[_T]]:
    edges: DefaultDict[_T, Set[_T]] = util.defaultdict(set)
    for parent, child in tuples:
        edges[child].add(parent)

    todo = list(allitems)
    todo_set = set(allitems)

    while todo_set:
        output = []
        for node in todo:
            if todo_set.isdisjoint(edges[node]):
                output.append(node)

        if not output:
            raise CircularDependencyError(
                "Circular dependency detected.",
                find_cycles(tuples, allitems),
                _gen_edges(edges),
            )

        todo_set.difference_update(output)
        todo = [t for t in todo if t in todo_set]
        yield output


def sort(
    tuples: Collection[Tuple[_T, _T]],
    allitems: Collection[_T],
    deterministic_order: bool = True,
) -> Iterator[_T]:
    """sort the given list of items by dependency.

    'tuples' is a list of tuples representing a partial ordering.

    deterministic_order is no longer used, the order is now always
    deterministic given the order of "allitems".    the flag is there
    for backwards compatibility with Alembic.

    """

    for set_ in sort_as_subsets(tuples, allitems):
        yield from set_


def find_cycles(
    tuples: Iterable[Tuple[_T, _T]], allitems: Iterable[_T]
) -> Set[_T]:
    # adapted from:
    # https://neopythonic.blogspot.com/2009/01/detecting-cycles-in-directed-graph.html

    edges: DefaultDict[_T, Set[_T]] = util.defaultdict(set)
    for parent, child in tuples:
        edges[parent].add(child)
    nodes_to_test = set(edges)

    output = set()

    # we'd like to find all nodes that are
    # involved in cycles, so we do the full
    # pass through the whole thing for each
    # node in the original list.

    # we can go just through parent edge nodes.
    # if a node is only a child and never a parent,
    # by definition it can't be part of a cycle.  same
    # if it's not in the edges at all.
    for node in nodes_to_test:
        stack = [node]
        todo = nodes_to_test.difference(stack)
        while stack:
            top = stack[-1]
            for node in edges[top]:
                if node in stack:
                    cyc = stack[stack.index(node) :]
                    todo.difference_update(cyc)
                    output.update(cyc)

                if node in todo:
                    stack.append(node)
                    todo.remove(node)
                    break
            else:
                node = stack.pop()
    return output


def _gen_edges(edges: DefaultDict[_T, Set[_T]]) -> Set[Tuple[_T, _T]]:
    return {(right, left) for left in edges for right in edges[left]}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: opentelemetry/proto/collector/metrics/v1/metrics_service.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from opentelemetry.proto.metrics.v1 import (
    metrics_pb2 as opentelemetry_dot_proto_dot_metrics_dot_v1_dot_metrics__pb2,
)


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n>opentelemetry/proto/collector/metrics/v1/metrics_service.proto\x12(opentelemetry.proto.collector.metrics.v1\x1a,opentelemetry/proto/metrics/v1/metrics.proto"h\n\x1b\x45xportMetricsServiceRequest\x12I\n\x10resource_metrics\x18\x01 \x03(\x0b\x32/.opentelemetry.proto.metrics.v1.ResourceMetrics"~\n\x1c\x45xportMetricsServiceResponse\x12^\n\x0fpartial_success\x18\x01 \x01(\x0b\x32\x45.opentelemetry.proto.collector.metrics.v1.ExportMetricsPartialSuccess"R\n\x1b\x45xportMetricsPartialSuccess\x12\x1c\n\x14rejected_data_points\x18\x01 \x01(\x03\x12\x15\n\rerror_message\x18\x02 \x01(\t2\xac\x01\n\x0eMetricsService\x12\x99\x01\n\x06\x45xport\x12\x45.opentelemetry.proto.collector.metrics.v1.ExportMetricsServiceRequest\x1a\x46.opentelemetry.proto.collector.metrics.v1.ExportMetricsServiceResponse"\x00\x42\xa4\x01\n+io.opentelemetry.proto.collector.metrics.v1B\x13MetricsServiceProtoP\x01Z3go.opentelemetry.io/proto/otlp/collector/metrics/v1\xaa\x02(OpenTelemetry.Proto.Collector.Metrics.V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR,
    "opentelemetry.proto.collector.metrics.v1.metrics_service_pb2",
    _globals,
)
if not _descriptor._USE_C_DESCRIPTORS:
    _globals["DESCRIPTOR"]._loaded_options = None
    _globals["DESCRIPTOR"]._serialized_options = (
        b"\n+io.opentelemetry.proto.collector.metrics.v1B\023MetricsServiceProtoP\001Z3go.opentelemetry.io/proto/otlp/collector/metrics/v1\252\002(OpenTelemetry.Proto.Collector.Metrics.V1"
    )
    _globals["_EXPORTMETRICSSERVICEREQUEST"]._serialized_start = 154
    _globals["_EXPORTMETRICSSERVICEREQUEST"]._serialized_end = 258
    _globals["_EXPORTMETRICSSERVICERESPONSE"]._serialized_start = 260
    _globals["_EXPORTMETRICSSERVICERESPONSE"]._serialized_end = 386
    _globals["_EXPORTMETRICSPARTIALSUCCESS"]._serialized_start = 388
    _globals["_EXPORTMETRICSPARTIALSUCCESS"]._serialized_end = 470
    _globals["_METRICSSERVICE"]._serialized_start = 473
    _globals["_METRICSSERVICE"]._serialized_end = 645
# @@protoc_insertion_point(module_scope)



import json

from order_api.dto.exception import (
    CannotSaveEmptyOrder,
    CannotSaveOrderWithNotFilledMandatoryOption,
)
from pos_api import PosService
from typing import (
    Dict,
    List,
    Optional,
    Union,
)
from xml.etree import cElementTree as eTree

from msgbus import FM_STRING
from messagebus import (
    ApplicationErrorMessage,
    DataType,
    DefaultToken,
    Message,
    MessageBus,
)
from .dto import (
    Table,
    TableMode,
    TableServiceDto,
    OrderLines,
)
from .dto.exception import (
    CurrentTabWithPartialTenders,
    InvalidTableStatus,
    InvalidTabMask,
    InvalidTabName,
    MissingAnyOfParameters,
    MissingNumSeats,
    OperatorNotLoggedIn,
    TabIsNotASplitSeat,
    TabNameAlreadyExists,
    TabNotFound,
    TabToJoinDoesNotExists,
    TabToJoinInInvalidStatus,
    TabToJoinIsAlreadyJoined,
    TabToJoinIsEmpty,
    TabToJoinIsTheSame,
    TabToJoinOpenedInAnotherPOS,
    TabToJoinWithPartialTenders,
    TableAlreadyOpened,
    TableIsNotInTabsMode,
    TableNotFound,
    TabNotInValidTabNames,
    TableNotOpened,
    TableServiceException,
    UnknownError,
)
from .dto.table_xml_parser import parse_table
from .error import CodedError
from .token import TK_CMP_MESSAGE


class TableService(object):

    def __init__(
        self,               # type: TableService
        message_bus,        # type: MessageBus
        pos_id,             # type: int
        pos_service=None,   # type: Optional[PosService]
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.pos_id = pos_id
        self.table_mgr = "TABLEMGR"

        self.pos_service = pos_service
        if self.pos_service is None:
            self.pos_service = PosService(
                message_bus=message_bus,
                pos_id=self.pos_id,
            )

    def get_table_picture(
        self,       # type: TableService
        table_id,   # type: int
    ):
        # type: (...) -> eTree.Element

        params = (
            "TABLE_PICTURE",
            self.pos_id,
            table_id,
        )
        request_data = "\x00".join((str(p) for p in params))
        msg = self.message_bus.send_message(
            component=self.table_mgr,
            message=Message(
                token=TK_CMP_MESSAGE,
                data_type=DataType.param,
                data=request_data.encode(),
                timeout=600000000,
            ),
        )
        if msg.token == DefaultToken.TK_SYS_ACK:
            return eTree.XML(msg.data)

        raise TableServiceException(description=msg.data.decode())

    def get_table(
        self,       # type: TableService
        table_id,   # type: int
    ):
        # type: (...) -> Table

        xml_table_picture = self.get_table_picture(table_id=table_id)
        table = parse_table(table_xml=xml_table_picture)
        return table

    def get_sectors(
        self,       # type: TableService
    ):
        # type: (...) -> List[str]

        reply = self.message_bus.send_message(
            component=self.table_mgr,
            message=Message(
                token=TK_CMP_MESSAGE,
                data=b"GET_SECTORS",
            )
        )

        if reply.token != DefaultToken.TK_SYS_ACK:
            raise TableServiceException(description=reply.data.decode())

        return json.loads(reply.data.decode())

    def start_service(
        self,                       # type: TableService
        table_id,                   # type: int
        mode,                       # type: TableMode
        num_seats=None,             # type: Optional[int]
        custom_properties=None,     # type: Optional[Dict[str, str]]
    ):
        # type: (...) -> TableServiceDto

        request = {
            "posId": self.pos_id,
            "tableId": table_id,
            "mode": mode.value,
        }
        if num_seats is not None:
            request["numSeats"] = num_seats

        if custom_properties is not None:
            request["customProperties"] = custom_properties

        params = (
            "START_SERVICE",
            json.dumps(request),
        )
        request_data = "\0".join((str(p) for p in params))
        reply = self.message_bus.send_message(
            component=self.table_mgr,
            message=Message(
                token=TK_CMP_MESSAGE,
                data_type=DataType.param,
                data=request_data.encode(),
                timeout=600000000,
            ),
        )

        if reply.token == DefaultToken.TK_SYS_APPLICATION_ERROR:
            app_error = ApplicationErrorMessage.build_from_message(message=reply)
            ex_dict = json.loads(app_error.error_message)
            if ex_dict["ex"] == "TableNotFound":
                raise TableNotFound(table_id=ex_dict["tableId"])
            elif ex_dict["ex"] == "OperatorNotLoggedIn":
                raise OperatorNotLoggedIn(pos_id=ex_dict["posId"])
            elif ex_dict["ex"] == "MissingNumSeats":
                raise MissingNumSeats()
            elif ex_dict["ex"] == "TableAlreadyOpened":
                raise TableAlreadyOpened(table_id=ex_dict["tableId"])

        if reply.token != DefaultToken.TK_SYS_ACK:
            raise UnknownError(
                token=reply.token,
                data=reply.data,
            )

        return self._build_table_service(reply)

    def start_tab(
        self,                       # type: TableService
        tab_name,                   # type: str
        parent_table_id=None,       # type: Optional[int]
        custom_properties=None,     # type: Optional[Dict[str, str]]
    ):
        request = {
            "posId": self.pos_id,
            "tabName": tab_name,
        }

        if parent_table_id is not None:
            request["parentTableId"] = parent_table_id

        if custom_properties is not None:
            request["customProperties"] = custom_properties

        params = (
            "START_TAB",
            json.dumps(request),
        )
        request_data = "\0".join((str(p) for p in params))
        reply = self.message_bus.send_message(
            component=self.table_mgr,
            message=Message(
                token=TK_CMP_MESSAGE,
                data_type=DataType.param,
                data=request_data.encode(),
                timeout=600000000,
            ),
        )

        if reply.token == DefaultToken.TK_SYS_APPLICATION_ERROR:
            app_error = ApplicationErrorMessage.build_from_message(message=reply)
            ex_dict = json.loads(app_error.error_message)
            if ex_dict["ex"] == "InvalidTabMask":
                raise InvalidTabMask(tab_name=ex_dict["tabName"])
            if ex_dict["ex"] == "TableNotFound":
                raise TableNotFound(table_id=ex_dict["tableId"])
            elif ex_dict["ex"] == "OperatorNotLoggedIn":
                raise OperatorNotLoggedIn(pos_id=ex_dict["posId"])
            elif ex_dict["ex"] == "MissingNumSeats":
                raise MissingNumSeats()
            elif ex_dict["ex"] == "TableAlreadyOpened":
                raise TableAlreadyOpened(table_id=ex_dict["tableId"])
            elif ex_dict["ex"] == "TabNameAlreadyExists":
                raise TabNameAlreadyExists(tab_name=ex_dict["tabName"])
            elif ex_dict["ex"] == "TableNotOpened":
                raise TableNotOpened(table_id=ex_dict["tableId"])
            elif ex_dict["ex"] == "TableIsNotInTabsMode":
                raise TableIsNotInTabsMode(table_id=ex_dict["tableId"])

        if reply.token != DefaultToken.TK_SYS_ACK:
            raise UnknownError(
                token=reply.token,
                data=reply.data,
            )

        return self._build_table_service(reply)

    def get_service(
        self,           # type: TableService
        table_id,       # type: Union[int, str]
    ):
        # type: (...) -> TableServiceDto

        request = {
            "tableId": table_id,
        }

        params = (
            "TABLE_SERVICE_PICTURE",
            json.dumps(request),
        )
        request_data = "\0".join((str(p) for p in params))
        reply = self.message_bus.send_message(
            component=self.table_mgr,
            message=Message(
                token=TK_CMP_MESSAGE,
                data_type=DataType.param,
                data=request_data.encode(),
                timeout=600000000,
            ),
        )

        return self._build_table_service(reply)

    def get_service_tab(
        self,           # type: TableService
        tab_id=None,    # type: Optional[str]
        tab_name=None,  # type: Optional[str]
        pos_id=None,    # type: Optional[int]
    ):
        # type: (...) -> TableServiceDto

        request = {}

        if tab_id:
            request["tabId"] = tab_id
        if tab_name:
            request["tabName"] = tab_name
        if pos_id:
            request["posId"] = pos_id

        if not request:
            raise MissingAnyOfParameters(params=["tab_id", "tab_name", "pos_id"])

        params = ["TAB_SERVICE_PICTURE", json.dumps(request)]
        message = Message(
            token=TK_CMP_MESSAGE,
            data_type=DataType.param,
            data="\0".join(params).encode(),
            timeout=600000000,
        )

        reply = self.message_bus.send_message(
            component=self.table_mgr,
            message=message,
        )

        if reply.token == DefaultToken.TK_SYS_APPLICATION_ERROR:
            app_error = ApplicationErrorMessage.build_from_message(message=reply)
            ex_dict = json.loads(app_error.error_message)
            if ex_dict["ex"] == "InvalidTabMask":
                raise InvalidTabMask(tab_name=ex_dict["tabName"])
            elif ex_dict["ex"] == "TabNotInValidTabNames":
                raise TabNotInValidTabNames(tab_name=ex_dict["tabName"])
            elif ex_dict["ex"] == "InvalidTabName":
                raise InvalidTabName(tab_name=ex_dict["tabName"])
            if ex_dict["ex"] == "TabNotFound":
                raise TabNotFound(
                    tab_id=ex_dict["tabId"],
                    tab_name=ex_dict["tabName"],
                    pos_id=ex_dict["posId"],
                )

            raise UnknownError(
                token=reply.token,
                data=reply.data,
            )
        if reply.token == DefaultToken.TK_SYS_BAD_REQUEST:
            ret = json.loads(reply.data)
            raise CodedError(
                code=ret["errorCode"],
                message=ret["errorMessage"],

            )

        return self._build_table_service(reply=reply)

    def create_tab_from_split_service(
        self,               # type: TableService
        table_id,           # type: str
        table_name,         # type: str
        properties=None,    # type: Optional[Dict[str, str]]
    ):
        # type: (...) -> TableServiceDto

        if not table_id or not table_name:
            raise MissingAnyOfParameters(params=["table_id", "table_name"])

        if not table_id.startswith("TAB-"):
            raise TabNotFound(tab_id=table_id)

        request = {
            "posId": self.pos_id,
            "tableId": table_id,
            "tableName": table_name,
            "properties": properties,
        }
        params = ("CREATE_TAB_FROM_SPLIT_SERVICE", json.dumps(request))
        reply = self.message_bus.send_message(
            component=self.table_mgr,
            message=Message(
                token=TK_CMP_MESSAGE,
                data_type=DataType.param,
                data="\0".join((str(p) for p in params)).encode(),
                timeout=600000000,
            ),
        )

        if reply.token == DefaultToken.TK_SYS_APPLICATION_ERROR:
            app_error = ApplicationErrorMessage.build_from_message(message=reply)
            ex_dict = json.loads(app_error.error_message)
            if ex_dict["ex"] == "InvalidTabMask":
                raise InvalidTabMask(tab_name=ex_dict["tabName"])
            elif ex_dict["ex"] == "TabNotFound":
                raise TabNotFound(tab_id=ex_dict["tabId"])
            elif ex_dict["ex"] == "TabNameAlreadyExists":
                raise TabNameAlreadyExists(tab_name=ex_dict["tabName"])
            elif ex_dict["ex"] == "InvalidTabName":
                raise InvalidTabName(tab_name=ex_dict["tabName"])
            elif ex_dict["ex"] == "TableNotOpened":
                raise TableNotOpened(table_id=ex_dict["tableId"])
            elif ex_dict["ex"] == "TabIsNotASplitSeat":
                raise TabIsNotASplitSeat(tab_name=ex_dict["tabName"])

        if reply.token != DefaultToken.TK_SYS_ACK:
            raise UnknownError(
                token=reply.token,
                data=reply.data,
            )

        return self._build_table_service(reply=reply)

    def join_tabs(
        self,               # type: TableService
        current_tab_id,     # type: str
        tab_name_to_join,   # type: str
    ):
        # type: (...) -> TableServiceDto

        if not current_tab_id or not tab_name_to_join:
            raise MissingAnyOfParameters(params=["current_tab_id", "tab_name_to_join"])

        user_id = str(self.pos_service.get_pos_model().logged_user.id)
        params = ["JOIN_TABS", str(self.pos_id), user_id, current_tab_id, tab_name_to_join]
        message = Message(
            token=TK_CMP_MESSAGE,
            data_type=DataType.param,
            data="\0".join(params).encode(),
            timeout=600000000,
        )

        reply = self.message_bus.send_message(
            component=self.table_mgr,
            message=message,
        )

        if reply.token == DefaultToken.TK_SYS_ACK:
            return self._build_table_service(reply=reply)

        self._tabs_error(ex_message=reply.data.decode())

    def unlink_tabs(
        self,               # type: TableService
        current_tab_id,     # type: str
        tab_ids_to_unlink,  # type: List[str]
    ):
        # type: (...) -> TableServiceDto

        if not current_tab_id or not tab_ids_to_unlink:
            raise MissingAnyOfParameters(params=["current_tab_id", "tab_ids_to_unlink"])

        user_id = str(self.pos_service.get_pos_model().logged_user.id)
        params = ["UNLINK_TABS", str(self.pos_id), user_id, current_tab_id, ",".join(tab_ids_to_unlink)]
        reply = self.message_bus.send_message(
            component=self.table_mgr,
            message=Message(
                token=TK_CMP_MESSAGE,
                data_type=DataType.param,
                data="\0".join(params).encode(),
                timeout=600000000,
            ),
        )

        if reply.token == DefaultToken.TK_SYS_ACK:
            return self._build_table_service(reply=reply)

        self._tabs_error(ex_message=reply.data.decode())

    def register_service_tender(
        self,                   # type: TableService
        table_id,               # type: str
        tender_id,              # type: int
        amount=None,            # type: Optional[float]
        tender_detail=None,     # type: Optional[Dict[str, Union[str, float, int]]]
    ):
        # type: (...) -> TableServiceDto

        if not table_id or tender_id is None:
            raise MissingAnyOfParameters(params=["table_id", "tender_id"])

        user_id = str(self.pos_service.get_pos_model().logged_user.id)
        amount = str(amount) if amount is not None else ""
        tender_id = str(tender_id)
        tender_details = json.dumps(tender_detail) if isinstance(tender_detail, dict) else ""
        params = ["REGISTER_SERVICE_TENDER", str(self.pos_id), user_id, table_id, tender_id, amount, tender_details]
        return self._process_old_call(params=params)

    def reopen_service(
        self,       # type: TableService
        table_id,   # type: str
    ):
        # type: (...) -> TableServiceDto

        if not table_id:
            raise MissingAnyOfParameters(params=["table_id"])

        user_id = str(self.pos_service.get_pos_model().logged_user.id)
        params = ["REOPEN_SERVICE", str(self.pos_id), user_id, table_id]
        return self._process_old_call(params=params)

    def recall_service(
        self,       # type: TableService
        table_id,   # type: str
    ):
        # type: (...) -> TableServiceDto

        if not table_id:
            raise MissingAnyOfParameters(params=["table_id"])

        user_id = str(self.pos_service.get_pos_model().logged_user.id)
        params = ["RECALL_SERVICE", str(self.pos_id), user_id, table_id]
        return self._process_old_call(params=params)

    def set_custom_property(
        self,               # type: TableService
        prop_key,           # type: str
        prop_value="",      # type: str
        table_id=None,      # type: Optional[str]
        service_id=None,    # type: Optional[int]
    ):
        # type: (...) -> None

        if not prop_key or (not table_id and not service_id):
            raise MissingAnyOfParameters(params=["prop_value", "table_id", "service_id"])

        params = ["SET_CUSTOM_PROPERTY", str(self.pos_id), prop_key, prop_value, table_id or "", service_id or ""]
        return self._process_old_call(params=params)

    def set_custom_properties(
        self,               # type: TableService
        properties=None,    # type: Union[str, Dict[str, Union[str, int, Dict, List]]]
        table_id=None,      # type: str
        service_id=None,    # type: str
    ):
        # type: (...) -> None

        if isinstance(properties, dict):
            properties = json.dumps(properties)

        params = ["SET_CUSTOM_PROPERTIES", str(self.pos_id), properties, table_id or "", service_id or ""]
        return self._process_old_call(params=params)

    def total_service(
        self,                   # type: TableService
        table_id,               # type: str
        tip_rate="",            # type: str
        tip_amount="",          # type: str
        sale_type="EAT_IN",     # type: str
    ):
        # type: (...) -> TableServiceDto

        if not table_id:
            raise MissingAnyOfParameters(params=["table_id"])

        user_id = str(self.pos_service.get_pos_model().logged_user.id)
        params = ["TOTAL_SERVICE", str(self.pos_id), user_id, table_id, tip_rate, tip_amount, sale_type]
        return self._process_old_call(params=params)

    def move_order_to_service(
        self,               # type: TableService
        order_id,           # type: int
        table_id=None,      # type: Optional[str]
        service_id=None,    # type: Optional[int]
    ):
        # type: (...) -> None

        if not order_id or (not table_id and not service_id):
            raise MissingAnyOfParameters(params=["order_id", "table_id", "service_id"])

        user_id = str(self.pos_service.get_pos_model().logged_user.id)
        service_id = str(service_id or "")
        params = ["MOVE_ORDER_TO_SERVICE", str(self.pos_id), user_id, table_id, str(order_id), service_id]
        self._process_old_call(params=params)

    def store_service(
        self,               # type: TableService
        table_id,           # type: str
        order_id=None,      # type: Optional[int]
    ):
        # type: (...) -> None

        if not table_id:
            raise MissingAnyOfParameters(params=["table_id"])

        user_id = str(self.pos_service.get_pos_model().logged_user.id)
        params = ["STORE_SERVICE", str(self.pos_id), user_id, table_id, str(order_id or "")]
        self._process_old_call(params=params)

    def close_service(
        self,       # type: TableService
        table_id,   # type: str
    ):
        # type: (...) -> TableServiceDto

        if not table_id:
            raise MissingAnyOfParameters(params=["table_id"])

        user_id = str(self.pos_service.get_pos_model().logged_user.id)
        params = ["CLOSE_SERVICE", str(self.pos_id), user_id, table_id]
        return self._process_old_call(params=params)

    def abandon_service(
        self,       # type: TableService
        table_id,   # type: str
    ):
        # type: (...) -> None

        if not table_id:
            raise MissingAnyOfParameters(params=["table_id"])

        user_id = str(self.pos_service.get_pos_model().logged_user.id)
        params = ["ABANDON_SERVICE", str(self.pos_id), user_id, table_id]
        return self._process_old_call(params=params)

    def move_tabs(
        self,              # type: TableService
        new_user_id,       # type: str
        from_tab_id,       # type: str
        dest_tab_id,       # type: str
        linked_tab,        # type: str
    ):
        # type: (...) -> Message

        params = [
            "MOVE_TABLE",
            str(self.pos_id),
            new_user_id,
            from_tab_id,
            dest_tab_id,
            "1",
            json.dumps([linked_tab]),
            "seats",
            "True",
            new_user_id,
            "None"
        ]

        msg = self.message_bus.send_message(
            component=self.table_mgr,
            message=Message(
                token=TK_CMP_MESSAGE,
                data_type=DataType.param,
                data="\0".join(params).encode(),
                timeout=600000000,
            ),
        )
        if msg.token == DefaultToken.TK_SYS_NAK:
            raise TableServiceException(
                errmsg=(msg.data if msg.format == FM_STRING and msg.data else "Error move tabs")
            )

        return msg

    def system_abandon_service(
        self,           # type: TableService
        table_id,       # type: str
    ):
        # type: (...) -> TableServiceDto

        if not table_id:
            raise MissingAnyOfParameters(params=["table_id"])

        user_id = 0
        params = ["ABANDON_SERVICE", str(self.pos_id), str(user_id), table_id]
        return self._process_old_call(params=params)

    def create_service_order(
        self,
        table_id,                   # type: str
        custom_properties=None,     # type: Dict[str, str]
        sale_type="EAT_IN",         # type: str
    ):
        # type: (...) -> int

        if not table_id:
            raise MissingAnyOfParameters(params=["table_id"])

        user_id = str(self.pos_service.get_pos_model().logged_user.id)
        properties = json.dumps(custom_properties) if custom_properties else ""
        params = ["CREATE_ORDER", str(self.pos_id), user_id, table_id, "", properties, sale_type]

        reply = self._get_reply(params=params)
        if reply.token == DefaultToken.TK_SYS_ACK:
            return int(eTree.XML(reply.data).get("orderId"))

        raise UnknownError(
            token=DefaultToken.TK_SYS_NAK,
            data=reply.data,
        )

    def change_number_of_seats(
        self,                   # type: TableService
        user_id,                # type: str
        seats,                  # type: str
        table_id,               # type: str
        service_id="",          # type: str
        custom_seats=None,      # type: Optional[List[str]]
    ):
        # type: (...) -> None

        params = [
            "CHANGE_NUMBER_OF_SEATS",
            str(self.pos_id),
            user_id,
            seats,
            table_id,
            service_id,
            ";".join(custom_seats) if custom_seats else "",
        ]

        reply = self._get_reply(params=params)
        if reply.token == DefaultToken.TK_SYS_ACK:
            return None

        raise UnknownError(
            token=DefaultToken.TK_SYS_NAK,
            data=reply.data,
        )

    def move_table(
        self,                           # type: TableService
        user_id,                        # type: str
        from_table_id,                  # type: str
        dest_table_id,                  # type: str
        number_of_seats="",             # type: str
        tables_to_set_available=None,   # type: Optional[List]
        mode="seats",                   # type: str
        ignore_totaled_orders=False,    # type: bool
        custom_seats=None,              # type: Optional[List[str]]
        current_user_id=None,           # type: Optional[Union[str, int]]
        authorizer=None,                # type: Optional[Union[str, int]]
    ):
        # type: (...) -> None

        params = [
            "MOVE_TABLE",
            str(self.pos_id),
            user_id,
            from_table_id,
            dest_table_id,
            number_of_seats,
            tables_to_set_available,
            mode,
            "true" if ignore_totaled_orders else "false",
            ";".join(custom_seats) if custom_seats else "",
            current_user_id,
            authorizer,
        ]

        reply = self._get_reply(params=params)
        if reply.token == DefaultToken.TK_SYS_ACK:
            return

        raise UnknownError(
            token=DefaultToken.TK_SYS_NAK,
            data=reply.data,
        )

    def move_table_order(
        self,                       # type: TableService
        user_id,                    # type: str
        from_table_id,              # type: str
        dest_table_id,              # type: str
        order_id,                   # type: int
        number_of_seats="",         # type: Union[str, int]
        custom_seats=None,          # type: Optional[List[str]]
        current_user_id=None,       # type: Optional[Union[str, int]]
        authorizer=None,            # type: Optional[Union[str, int]]
    ):
        # type: (...) -> None

        params = [
            "MOVE_TABLE_ORDER",
            str(self.pos_id),
            user_id,
            from_table_id,
            dest_table_id,
            str(order_id),
            number_of_seats,
            ";".join(custom_seats) if custom_seats else "",
            current_user_id,
            authorizer
        ]

        reply = self._get_reply(params=params)
        if reply.token == DefaultToken.TK_SYS_ACK:
            return

        if "INVALID_TABLE_STATUS" in reply.data:
            raise InvalidTableStatus()

        raise UnknownError(
            token=DefaultToken.TK_SYS_NAK,
            data=reply.data,
        )

    def join_tables(
        self,                       # type: TableService
        user_id,                    # type: str
        from_table_id,              # type: str
        dest_table_id,              # type: str
        number_of_seats="",         # type: str
        custom_seats=None,          # type: Optional[List[str]]
        current_user_id=None,       # type: Optional[Union[str, int]]
        authorizer=None,            # type: Optional[Union[str, int]]
    ):
        # type: (...) -> None

        params = [
            "JOIN_TABLES",
            str(self.pos_id),
            user_id,
            from_table_id,
            dest_table_id,
            number_of_seats,
            ";".join(custom_seats) if custom_seats else "",
            current_user_id,
            authorizer,
        ]

        reply = self._get_reply(params=params)
        if reply.token == DefaultToken.TK_SYS_ACK:
            return

        if "INVALID_TABLE_STATUS" in reply.data:
            raise InvalidTableStatus()

        raise UnknownError(
            token=DefaultToken.TK_SYS_NAK,
            data=reply.data,
        )

    def change_seat_number(
        self,               # type: TableService
        table_id,           # type: str
        old_seat_number,    # type: str
        new_seat_number,    # type: str
    ):
        # type: (...) -> None

        params = [
            "CHANGE_SEAT_NUMBER",
            str(self.pos_id),
            table_id,
            old_seat_number,
            new_seat_number,
        ]

        reply = self._get_reply(params=params)
        if reply.token == DefaultToken.TK_SYS_ACK:
            return

        raise UnknownError(
            token=DefaultToken.TK_SYS_NAK,
            data=reply.data,
        )

    def change_sale_lines_seats(
        self,           # type: TableService
        table_id,       # type: str
        to_seat,        # type: str
        orders_lines,   # type: List[OrderLines]
    ):
        # type: (...) -> str

        params = [
            "CHANGE_SALE_LINES_SEATS",
            str(self.pos_id),
            table_id,
            to_seat,
            str(orders_lines),
        ]

        reply = self._get_reply(params=params)
        if reply.token == DefaultToken.TK_SYS_ACK:
            return reply.data

        raise UnknownError(
            token=DefaultToken.TK_SYS_NAK,
            data=reply.data,
        )

    def _process_old_call(
        self,       # type: TableService
        params,     # type: List[str]
    ):
        # type: (...) -> Optional[TableServiceDto]

        reply = self._get_reply(params=params)

        if reply.token == DefaultToken.TK_SYS_ACK:
            return self._build_table_service(reply=reply) if reply.data else None

        if reply.data == CannotSaveEmptyOrder.get_error_code():
            raise CannotSaveEmptyOrder()

        if "|" in reply.data:
            splitted_data = reply.data.decode("utf-8").split("|")
            if splitted_data[0] == CannotSaveOrderWithNotFilledMandatoryOption.get_error_code():
                raise CannotSaveOrderWithNotFilledMandatoryOption(
                    part_code=int(splitted_data[1]),
                    part_name=splitted_data[2],
                )

        raise UnknownError(
            token=DefaultToken.TK_SYS_NAK,
            data=reply.data,
        )

    def _get_reply(
        self,       # type: TableService
        params,     # type: List[str]
    ):
        # type: (...) -> Message

        return self.message_bus.send_message(
            component=self.table_mgr,
            message=Message(
                token=TK_CMP_MESSAGE,
                data_type=DataType.param,
                data="\0".join(params).encode(),
                timeout=600000000,
            ),
        )

    @staticmethod
    def _tabs_error(
        ex_message,     # type: str
    ):
        # type: (...) -> None

        ex_params = str(ex_message).split("|")

        if "$TAB_NOT_FOUND" == ex_params[0]:
            raise InvalidTabMask(tab_name=ex_params[1])

        if "$TAB_NOT_REGISTERED" == ex_params[0]:
            raise TabNotInValidTabNames(tab_name=ex_params[1])

        if "$SELECT_A_DIFFERENT_TAB" == ex_params[0]:
            raise TabToJoinIsTheSame()

        if "$TAB_TO_JOIN_DOES_NOT_EXISTS" == ex_params[0]:
            raise TabToJoinDoesNotExists(tab_name=ex_params[1])

        if "$TAB_TO_JOIN_IS_EMPTY" == ex_params[0]:
            raise TabToJoinIsEmpty(tab_name=ex_params[1])

        if "$TAB_TO_JOIN_IS_ALREADY_JOINED" == ex_params[0]:
            raise TabToJoinIsAlreadyJoined(tab_name=ex_params[1])

        if "$TAB_TO_JOIN_IS_OPENED_IN_ANOTHER_POS" == ex_params[0]:
            raise TabToJoinOpenedInAnotherPOS(tab_name=ex_params[1])

        if "$CURRENT_TAB_WITH_PARTIAL_TENDERS" == ex_params[0]:
            raise CurrentTabWithPartialTenders(table_id=ex_params[1])

        if "$TAB_TO_JOIN_WITH_PARTIAL_TENDERS" == ex_params[0]:
            raise TabToJoinWithPartialTenders(table_id=ex_params[1])

        if "$TAB_TO_JOIN_IN_INVALID_STATUS" == ex_params[0]:
            ex_params = ex_params[ex_params.index("$TAB_TO_JOIN_IN_INVALID_STATUS"):]
            raise TabToJoinInInvalidStatus(
                state=ex_params[2],
                tab_name=ex_params[1],
            )

        raise UnknownError(
            token=DefaultToken.TK_SYS_NAK,
            data=ex_message.encode(),
        )

    @staticmethod
    def _build_table_service(
        reply,      # type: Message
    ):
        # type: (...) -> TableServiceDto

        reply_xml = eTree.XML(reply.data)
        return TableServiceDto.from_xml(xml=reply_xml)



import threading


class ScaleEvent(object):
    def __init__(self):
        self.event = threading.Event()
        self.weight = None

    def get_weight(self):
        return self.weight

    def set_weight(self, weight):
        self.weight = weight
        self.event.set()

    def clear_weight(self):
        self.weight = None
        self.event.clear()

    def wait(self, timeout):
        self.event.wait(timeout)
        self.event.clear()



from typing import Optional

from boss.interactor.get_info.dto import GetInfoRequest
from boss.model import User
from boss.util import convert_user_dto_into_user_response
from boss_api.shared.exception import (
    NoData,
    NoUser,
    NoUtf8Data,
)
from messagebus import (
    AckMessage,
    Event,
    Message,
    MessageBus,
)
from ._BaseProcessor import BaseProcessor


class GetInfoProcessor(BaseProcessor):

    def parse_data(
        self,
        data: bytes,
    ) -> GetInfoRequest:

        if not data:
            raise NoData()

        try:
            data.decode()
        except UnicodeDecodeError:
            raise NoUtf8Data()

        try:
            user_id = int(data.decode())
        except ValueError:
            raise NoUser()

        if not user_id:
            raise NoUser()

        return GetInfoRequest(user_id=user_id)

    def format_response(
        self,
        message_bus: MessageBus,
        message: Message,
        event: Event,
        request: GetInfoRequest,
        result: Optional[User],
    ) -> None:

        if result is None:
            reply_message = AckMessage(data=b"<UserInfo/>")
        else:
            user = convert_user_dto_into_user_response(user=result)
            reply_data = f"<UserInfo>{user.get_info_response()}</UserInfo>"
            reply_message = AckMessage(data=reply_data.encode())

        message_bus.reply_message(
            message=message,
            reply_message=reply_message,
        )



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Final

from deprecated import deprecated

CONTAINER_COMMAND: Final = "container.command"
"""
The command used to run the container (i.e. the command name).
Note: If using embedded credentials or sensitive data, it is recommended to remove them to prevent potential leakage.
"""

CONTAINER_COMMAND_ARGS: Final = "container.command_args"
"""
All the command arguments (including the command/executable itself) run by the container.
"""

CONTAINER_COMMAND_LINE: Final = "container.command_line"
"""
The full command run by the container as a single string representing the full command.
"""

CONTAINER_CPU_STATE: Final = "container.cpu.state"
"""
Deprecated: Replaced by `cpu.mode`.
"""

CONTAINER_CSI_PLUGIN_NAME: Final = "container.csi.plugin.name"
"""
The name of the CSI ([Container Storage Interface](https://github.com/container-storage-interface/spec)) plugin used by the volume.
Note: This can sometimes be referred to as a "driver" in CSI implementations. This should represent the `name` field of the GetPluginInfo RPC.
"""

CONTAINER_CSI_VOLUME_ID: Final = "container.csi.volume.id"
"""
The unique volume ID returned by the CSI ([Container Storage Interface](https://github.com/container-storage-interface/spec)) plugin.
Note: This can sometimes be referred to as a "volume handle" in CSI implementations. This should represent the `Volume.volume_id` field in CSI spec.
"""

CONTAINER_ID: Final = "container.id"
"""
Container ID. Usually a UUID, as for example used to [identify Docker containers](https://docs.docker.com/engine/containers/run/#container-identification). The UUID might be abbreviated.
"""

CONTAINER_IMAGE_ID: Final = "container.image.id"
"""
Runtime specific image identifier. Usually a hash algorithm followed by a UUID.
Note: Docker defines a sha256 of the image id; `container.image.id` corresponds to the `Image` field from the Docker container inspect [API](https://docs.docker.com/engine/api/v1.43/#tag/Container/operation/ContainerInspect) endpoint.
K8s defines a link to the container registry repository with digest `"imageID": "registry.azurecr.io /namespace/service/dockerfile@sha256:bdeabd40c3a8a492eaf9e8e44d0ebbb84bac7ee25ac0cf8a7159d25f62555625"`.
The ID is assigned by the container runtime and can vary in different environments. Consider using `oci.manifest.digest` if it is important to identify the same image in different environments/runtimes.
"""

CONTAINER_IMAGE_NAME: Final = "container.image.name"
"""
Name of the image the container was built on.
"""

CONTAINER_IMAGE_REPO_DIGESTS: Final = "container.image.repo_digests"
"""
Repo digests of the container image as provided by the container runtime.
Note: [Docker](https://docs.docker.com/engine/api/v1.43/#tag/Image/operation/ImageInspect) and [CRI](https://github.com/kubernetes/cri-api/blob/c75ef5b473bbe2d0a4fc92f82235efd665ea8e9f/pkg/apis/runtime/v1/api.proto#L1237-L1238) report those under the `RepoDigests` field.
"""

CONTAINER_IMAGE_TAGS: Final = "container.image.tags"
"""
Container image tags. An example can be found in [Docker Image Inspect](https://docs.docker.com/engine/api/v1.43/#tag/Image/operation/ImageInspect). Should be only the `<tag>` section of the full name for example from `registry.example.com/my-org/my-image:<tag>`.
"""

CONTAINER_LABEL_TEMPLATE: Final = "container.label"
"""
Container labels, `<key>` being the label name, the value being the label value.
"""

CONTAINER_LABELS_TEMPLATE: Final = "container.labels"
"""
Deprecated: Replaced by `container.label`.
"""

CONTAINER_NAME: Final = "container.name"
"""
Container name used by container runtime.
"""

CONTAINER_RUNTIME: Final = "container.runtime"
"""
The container runtime managing this container.
"""


@deprecated(
    reason="The attribute container.cpu.state is deprecated - Replaced by `cpu.mode`"
)  # type: ignore
class ContainerCpuStateValues(Enum):
    USER = "user"
    """When tasks of the cgroup are in user mode (Linux). When all container processes are in user mode (Windows)."""
    SYSTEM = "system"
    """When CPU is used by the system (host OS)."""
    KERNEL = "kernel"
    """When tasks of the cgroup are in kernel mode (Linux). When all container processes are in kernel mode (Windows)."""



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Final

from opentelemetry.metrics import Histogram, Meter

RPC_CLIENT_DURATION: Final = "rpc.client.duration"
"""
Measures the duration of outbound RPC
Instrument: histogram
Unit: ms
Note: While streaming RPCs may record this metric as start-of-batch
to end-of-batch, it's hard to interpret in practice.

**Streaming**: N/A.
"""


def create_rpc_client_duration(meter: Meter) -> Histogram:
    """Measures the duration of outbound RPC"""
    return meter.create_histogram(
        name=RPC_CLIENT_DURATION,
        description="Measures the duration of outbound RPC.",
        unit="ms",
    )


RPC_CLIENT_REQUEST_SIZE: Final = "rpc.client.request.size"
"""
Measures the size of RPC request messages (uncompressed)
Instrument: histogram
Unit: By
Note: **Streaming**: Recorded per message in a streaming batch.
"""


def create_rpc_client_request_size(meter: Meter) -> Histogram:
    """Measures the size of RPC request messages (uncompressed)"""
    return meter.create_histogram(
        name=RPC_CLIENT_REQUEST_SIZE,
        description="Measures the size of RPC request messages (uncompressed).",
        unit="By",
    )


RPC_CLIENT_REQUESTS_PER_RPC: Final = "rpc.client.requests_per_rpc"
"""
Measures the number of messages received per RPC
Instrument: histogram
Unit: {count}
Note: Should be 1 for all non-streaming RPCs.

**Streaming**: This metric is required for server and client streaming RPCs.
"""


def create_rpc_client_requests_per_rpc(meter: Meter) -> Histogram:
    """Measures the number of messages received per RPC"""
    return meter.create_histogram(
        name=RPC_CLIENT_REQUESTS_PER_RPC,
        description="Measures the number of messages received per RPC.",
        unit="{count}",
    )


RPC_CLIENT_RESPONSE_SIZE: Final = "rpc.client.response.size"
"""
Measures the size of RPC response messages (uncompressed)
Instrument: histogram
Unit: By
Note: **Streaming**: Recorded per response in a streaming batch.
"""


def create_rpc_client_response_size(meter: Meter) -> Histogram:
    """Measures the size of RPC response messages (uncompressed)"""
    return meter.create_histogram(
        name=RPC_CLIENT_RESPONSE_SIZE,
        description="Measures the size of RPC response messages (uncompressed).",
        unit="By",
    )


RPC_CLIENT_RESPONSES_PER_RPC: Final = "rpc.client.responses_per_rpc"
"""
Measures the number of messages sent per RPC
Instrument: histogram
Unit: {count}
Note: Should be 1 for all non-streaming RPCs.

**Streaming**: This metric is required for server and client streaming RPCs.
"""


def create_rpc_client_responses_per_rpc(meter: Meter) -> Histogram:
    """Measures the number of messages sent per RPC"""
    return meter.create_histogram(
        name=RPC_CLIENT_RESPONSES_PER_RPC,
        description="Measures the number of messages sent per RPC.",
        unit="{count}",
    )


RPC_SERVER_DURATION: Final = "rpc.server.duration"
"""
Measures the duration of inbound RPC
Instrument: histogram
Unit: ms
Note: While streaming RPCs may record this metric as start-of-batch
to end-of-batch, it's hard to interpret in practice.

**Streaming**: N/A.
"""


def create_rpc_server_duration(meter: Meter) -> Histogram:
    """Measures the duration of inbound RPC"""
    return meter.create_histogram(
        name=RPC_SERVER_DURATION,
        description="Measures the duration of inbound RPC.",
        unit="ms",
    )


RPC_SERVER_REQUEST_SIZE: Final = "rpc.server.request.size"
"""
Measures the size of RPC request messages (uncompressed)
Instrument: histogram
Unit: By
Note: **Streaming**: Recorded per message in a streaming batch.
"""


def create_rpc_server_request_size(meter: Meter) -> Histogram:
    """Measures the size of RPC request messages (uncompressed)"""
    return meter.create_histogram(
        name=RPC_SERVER_REQUEST_SIZE,
        description="Measures the size of RPC request messages (uncompressed).",
        unit="By",
    )


RPC_SERVER_REQUESTS_PER_RPC: Final = "rpc.server.requests_per_rpc"
"""
Measures the number of messages received per RPC
Instrument: histogram
Unit: {count}
Note: Should be 1 for all non-streaming RPCs.

**Streaming** : This metric is required for server and client streaming RPCs.
"""


def create_rpc_server_requests_per_rpc(meter: Meter) -> Histogram:
    """Measures the number of messages received per RPC"""
    return meter.create_histogram(
        name=RPC_SERVER_REQUESTS_PER_RPC,
        description="Measures the number of messages received per RPC.",
        unit="{count}",
    )


RPC_SERVER_RESPONSE_SIZE: Final = "rpc.server.response.size"
"""
Measures the size of RPC response messages (uncompressed)
Instrument: histogram
Unit: By
Note: **Streaming**: Recorded per response in a streaming batch.
"""


def create_rpc_server_response_size(meter: Meter) -> Histogram:
    """Measures the size of RPC response messages (uncompressed)"""
    return meter.create_histogram(
        name=RPC_SERVER_RESPONSE_SIZE,
        description="Measures the size of RPC response messages (uncompressed).",
        unit="By",
    )


RPC_SERVER_RESPONSES_PER_RPC: Final = "rpc.server.responses_per_rpc"
"""
Measures the number of messages sent per RPC
Instrument: histogram
Unit: {count}
Note: Should be 1 for all non-streaming RPCs.

**Streaming**: This metric is required for server and client streaming RPCs.
"""


def create_rpc_server_responses_per_rpc(meter: Meter) -> Histogram:
    """Measures the number of messages sent per RPC"""
    return meter.create_histogram(
        name=RPC_SERVER_RESPONSES_PER_RPC,
        description="Measures the number of messages sent per RPC.",
        unit="{count}",
    )



from ._I18N import I18N
from ._I18NRepository import I18NRepository
from ._L10NRepository import L10NRepository
from ._OrderAPIRepository import OrderAPIRepository
from ._POSControllerRepository import POSControllerRepository
from ._ProductRepository import ProductRepository
from ._StoreConfigRepository import StoreConfigRepository
from ._UserRepository import UserRepository



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

from opentelemetry.context import Context
from opentelemetry.util.types import Attributes


class Observation:
    """A measurement observed in an asynchronous instrument

    Return/yield instances of this class from asynchronous instrument callbacks.

    Args:
        value: The float or int measured value
        attributes: The measurement's attributes
        context: The measurement's context
    """

    def __init__(
        self,
        value: Union[int, float],
        attributes: Attributes = None,
        context: Optional[Context] = None,
    ) -> None:
        self._value = value
        self._attributes = attributes
        self._context = context

    @property
    def value(self) -> Union[float, int]:
        return self._value

    @property
    def attributes(self) -> Attributes:
        return self._attributes

    @property
    def context(self) -> Optional[Context]:
        return self._context

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Observation)
            and self.value == other.value
            and self.attributes == other.attributes
            and self.context == other.context
        )

    def __repr__(self) -> str:
        return f"Observation(value={self.value}, attributes={self.attributes}, context={self.context})"



import unittest

import importlib_resources as resources
from . import util


class CommonBinaryTests(util.CommonTests, unittest.TestCase):
    def execute(self, package, path):
        target = resources.files(package).joinpath(path)
        with target.open('rb'):
            pass


class CommonTextTests(util.CommonTests, unittest.TestCase):
    def execute(self, package, path):
        target = resources.files(package).joinpath(path)
        with target.open(encoding='utf-8'):
            pass


class OpenTests:
    def test_open_binary(self):
        target = resources.files(self.data) / 'binary.file'
        with target.open('rb') as fp:
            result = fp.read()
            self.assertEqual(result, bytes(range(4)))

    def test_open_text_default_encoding(self):
        target = resources.files(self.data) / 'utf-8.file'
        with target.open(encoding='utf-8') as fp:
            result = fp.read()
            self.assertEqual(result, 'Hello, UTF-8 world!\n')

    def test_open_text_given_encoding(self):
        target = resources.files(self.data) / 'utf-16.file'
        with target.open(encoding='utf-16', errors='strict') as fp:
            result = fp.read()
        self.assertEqual(result, 'Hello, UTF-16 world!\n')

    def test_open_text_with_errors(self):
        """
        Raises UnicodeError without the 'errors' argument.
        """
        target = resources.files(self.data) / 'utf-16.file'
        with target.open(encoding='utf-8', errors='strict') as fp:
            self.assertRaises(UnicodeError, fp.read)
        with target.open(encoding='utf-8', errors='ignore') as fp:
            result = fp.read()
        self.assertEqual(
            result,
            'H\x00e\x00l\x00l\x00o\x00,\x00 '
            '\x00U\x00T\x00F\x00-\x001\x006\x00 '
            '\x00w\x00o\x00r\x00l\x00d\x00!\x00\n\x00',
        )

    def test_open_binary_FileNotFoundError(self):
        target = resources.files(self.data) / 'does-not-exist'
        with self.assertRaises(FileNotFoundError):
            target.open('rb')

    def test_open_text_FileNotFoundError(self):
        target = resources.files(self.data) / 'does-not-exist'
        with self.assertRaises(FileNotFoundError):
            target.open(encoding='utf-8')


class OpenDiskTests(OpenTests, util.DiskSetup, unittest.TestCase):
    pass


class OpenDiskNamespaceTests(OpenTests, util.DiskSetup, unittest.TestCase):
    MODULE = 'namespacedata01'


class OpenZipTests(OpenTests, util.ZipSetup, unittest.TestCase):
    pass


class OpenNamespaceZipTests(OpenTests, util.ZipSetup, unittest.TestCase):
    MODULE = 'namespacedata01'


if __name__ == '__main__':
    unittest.main()



from typing import (
    Dict,
    Union,
)


class HTMLTextFormatter(object):

    def __init__(
        self,             # type: HTMLTextFormatter
        customization,    # type: Dict[str, Union[str, bool, int, float]]
        columns,          # type: int
    ):
        # type: (...) -> None

        self.columns = columns
        self.customization = customization
        self.style = {
            "font-family": "monospace",
            "font-size": "1.3em",
        }

    def format(
        self,    # type: HTMLTextFormatter
        text,    # type: str
    ):
        # type: (...) -> str

        if self.customization is None:
            return self._make_tag(text=text)

        formatted_text = self._handle_align(text=text)
        self._handle_bold()

        return self._make_tag(text=formatted_text)

    def _handle_bold(
        self,
    ):
        # type: (...) -> None

        if self.customization.get("bold"):
            self.style["font-weight"] = "bold"

    def _make_tag(
        self,    # type: HTMLTextFormatter
        text,    # type: str
    ):
        # type: (...) -> str

        class_attributes = ";".join(["{}: {}".format(key, value) for key, value in self.style.items()])

        return u"<span style='{}'>{}</span>".format(
            class_attributes,
            text.replace(" ", "&nbsp;"),
        )

    def _handle_align(
        self,    # type: HTMLTextFormatter
        text,    # type: str
    ):
        # type: (...) -> str

        formatted_text = text
        align = self.customization.get("align", "center")
        diff = self.columns - len(text)
        if align == "left":
            formatted_text = (" " * diff) + text
        elif align == "right":
            formatted_text = text + (" " * diff)
        elif align == "center":
            half = int(diff / 2)
            formatted_text = (" " * half) + text + (" " * half)

        return formatted_text



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from typing import Optional, Union

from opentelemetry.util.types import Attributes


@dataclasses.dataclass(frozen=True)
class Exemplar:
    """A representation of an exemplar, which is a sample input measurement.

    Exemplars also hold information about the environment when the measurement
    was recorded, for example the span and trace ID of the active span when the
    exemplar was recorded.

    Attributes
        trace_id: (optional) The trace associated with a recording
        span_id: (optional) The span associated with a recording
        time_unix_nano: The time of the observation
        value: The recorded value
        filtered_attributes: A set of filtered attributes which provide additional insight into the Context when the observation was made.

    References:
        https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/metrics/data-model.md#exemplars
        https://github.com/open-telemetry/opentelemetry-specification/blob/main/specification/metrics/sdk.md#exemplar
    """

    # TODO Fix doc - if using valid Google `Attributes:` key, the attributes are duplicated
    # one will come from napoleon extension and the other from autodoc extension. This
    # will raise an sphinx error of duplicated object description
    # See https://github.com/sphinx-doc/sphinx/issues/8664

    filtered_attributes: Attributes
    value: Union[int, float]
    time_unix_nano: int
    span_id: Optional[int] = None
    trace_id: Optional[int] = None



from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder
from production.box import AddAllItemsTagBox


class AddAllItemsTagBoxBuilder(BoxBuilder):
    def __init__(self, order_changer, loggers):
        super(AddAllItemsTagBoxBuilder, self).__init__(loggers)
        self.order_changer = order_changer

    def build(
        self,
        config,     # type: BoxConfiguration
    ):
        # type: (...) -> AddAllItemsTagBox

        return AddAllItemsTagBox(
            name=config.name,
            sons=config.sons,
            tag=self.get_extra(config, "Tag", None),
            order_changer=self.order_changer,
            logger=self.get_logger(config)
        )



from __future__ import annotations

import sys

if sys.version_info >= (3, 10):
    from typing import TypeGuard
else:
    try:
        from typing_extensions import TypeGuard
    except ImportError:
        from typing import Any

        class TypeGuard:  # type: ignore[no-redef]
            def __class_getitem__(cls, item: Any) -> type[bool]:
                return bool


__all__ = ["TypeGuard"]



"""

Revision ID: 37f226415b7f
Revises: 2f33894eaa09
Create Date: 2024-11-27 10:28:21.831752

"""
from typing import (
    Sequence,
    Union,
)

from alembic import op
from sqlalchemy import (
    Column,
    DATETIME,
    TEXT,
    VARCHAR,
)

# revision identifiers, used by Alembic.
revision: str = "37f226415b7f"
down_revision: Union[str, None] = "2f33894eaa09"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    op.drop_column("Users", "PasswordValidity")
    op.drop_column("Users", "PayRate")
    op.drop_column("Users", "BadgeNumber")


def downgrade() -> None:

    op.add_column("Users", Column("PasswordValidity", DATETIME))
    op.add_column("Users", Column("PayRate", VARCHAR(30)))
    op.add_column("Users", Column("BadgeNumber", TEXT))



from production.command import ProductionCommand


class PublishOrderCommand(ProductionCommand):
    def __init__(self, order_id):
        super(PublishOrderCommand, self).__init__(order_id, None)

    def get_hash_extra_value(self):
        return "PublishOrderCommand"
